/******************************************************************************
  Copyright (c) 1992 Xerox Corporation.  All rights reserved.
  Portions of this code were written by Stephen White, aka ghond.
  Use and copying of this software and preparation of derivative works based
  upon this software are permitted.  Any distribution of this software or
  derivative works must comply with all applicable United States export
  control laws.  This software is made available AS IS, and Xerox Corporation
  makes no warranty about the software, its performance or its conformity to
  any specification.  Any person obtaining a copy of this software is requested
  to send their name and post office or electronic mail address to:
    Pavel Curtis
    Xerox PARC
    3333 Coyote Hill Rd.
    Palo Alto, CA 94304
    Pavel@Xerox.Com
 *****************************************************************************/

#include "my-stdio.h"
#include "my-stdlib.h"
#include "my-types.h"
#include "my-unistd.h"

#include "config.h"
#include "db.h"
#include "db_property.h"
#include "exceptions.h"
#include "functions.h"
#include "list.h"
#include "log.h"
#include "quota.h"
#include "server.h"
#include "storage.h"
#include "streams.h"
#include "structures.h"
#include "timers.h"
#include "utils.h"

typedef enum { DUMP_SHUTDOWN, DUMP_CHECKPOINT, DUMP_PANIC } Dump_Reason;
const char *reason_names[] = { "DUMPING", "CHECKPOINTING", "PANIC-DUMPING" };

static char *input_db, *dump_db_name;
static int		dump_generation = 0;

/**** db ****/
int 
db_init(const char *in, const char *out)
{
    FILE       *f;
    
    input_db = str_dup(in);
    dump_db_name = str_dup(out);
    
    register_bi_functions();

    logf("LOADING: %s\n", input_db);

    if ((f = fopen(input_db, "r")) == 0) {
	errlogf("Cannot open database file: %s\n", input_db);
	return 1;
    }
    
    if (db_read(f) < 0) {
	errlogf("Cannot parse database file!\n");
	return 1;
    }
    
    logf("LOADING: %s done, will dump new database on %s\n", input_db, dump_db_name);

    fclose(f);
    return 0;
}

static int
dump_database(Dump_Reason reason)
{
    Stream     *s = new_stream(100);
    char       *temp_name;
    FILE       *f;
    int		success;

  retryDumping:

    stream_printf(s, "%s.#%d#", dump_db_name, dump_generation);
    remove(reset_stream(s));	/* Remove previous checkpoint */

    if (reason == DUMP_PANIC)
	stream_printf(s, "%s.PANIC", dump_db_name);
    else {
	dump_generation++;
	stream_printf(s, "%s.#%d#", dump_db_name, dump_generation);
    }
    temp_name = reset_stream(s);

    logf("%s on %s ...\n", reason_names[reason], temp_name);

#ifdef UNFORKED_CHECKPOINTS
    reset_command_history();
#else
    if (reason == DUMP_CHECKPOINT) {
	pid_t	result = fork();

	if (result == 0) {
	    in_child = 1;
	    set_server_cmdline("(MOO checkpointer)");
	} else {		/* in parent, possibly with an error */
	    if (result < 0)
		log_perror("Forking checkpointer");
	    else
		reset_command_history();
	    free_stream(s);
	    return result > 0;
	}
    }
#endif

    success = 1;
    if ((f = fopen(temp_name, "w")) != 0) {
	if (!db_write(f)) {
	    log_perror("Trying to dump database");
	    fclose(f);
	    remove(temp_name);
	    if (reason == DUMP_CHECKPOINT) {
		errlogf("Abandoning checkpoint attempt...\n");
		success = 0;
	    } else {
		int	retry_interval = 60;

		errlogf("Waiting %d seconds and retrying dump...\n",
		       retry_interval);
		timer_sleep(retry_interval);
		goto retryDumping;
	    }
	} else {
	    fclose(f);
	    logf("%s on %s finished\n", reason_names[reason], temp_name);
	    if (reason != DUMP_PANIC) {
		remove(dump_db_name);
		if (rename(temp_name, dump_db_name) != 0) {
		    log_perror("Renaming temporary dump file");
		    success = 0;
		}
	    }
	}
    } else {
	log_perror("Opening temporary dump file");
	success = 0;
    }

    free_stream(s);

#ifndef UNFORKED_CHECKPOINTS
    if (reason == DUMP_CHECKPOINT)
	/* We're a child, so we'd better go away. */
	exit(!success);
#endif

    return success;
}

int 
db_flush(enum flush_type type)
{
    int	success = 0;

    switch(type) {
      case FLUSH_IF_FULL:
      case FLUSH_ONE_SECOND:
	success = 1;
	break;

      case FLUSH_ALL_NOW:
	success = dump_database(DUMP_CHECKPOINT);
	break;

      case FLUSH_PANIC:
	success = dump_database(DUMP_PANIC);
	break;
    }

    return success;
}

void
db_shutdown()
{
    dump_database(DUMP_SHUTDOWN);
}

/**** objects ****/
void
add_child(Objid parent, Objid oid)
{
    Objid lid; 
    if (obj_child(parent) == NOTHING) {
	set_obj_child(parent, oid);
    } else { 
	for (lid = obj_child(parent);
	     obj_sibling(lid) != NOTHING; 
	     lid = obj_sibling(lid)) 
	    ; 
	set_obj_sibling(lid, oid);
    }
    set_obj_sibling(oid, NOTHING);
}

void
remove_child(Objid parent, Objid oid)
{
    Objid lid; 
    if (obj_child(parent) == oid) 
	set_obj_child(parent, obj_sibling(oid));
    else { 
	for (lid = obj_child(parent); lid != NOTHING;
	     lid = obj_sibling(lid)) {
	    if (obj_sibling(lid) == oid) { 
		set_obj_sibling(lid, obj_sibling(oid));
		break; 
	    } 
	} 
    } 
    set_obj_sibling(oid, NOTHING);
}

void
add_content(Objid where, Objid oid)
{
    Objid lid; 
    if (obj_contents(where) == NOTHING) {
	set_obj_contents(where, oid);
    } else { 
	for (lid = obj_contents(where);
	     obj_next_content(lid) != NOTHING; 
	     lid = obj_next_content(lid)) 
	    ; 
	set_obj_next_content(lid, oid);
    }
    set_obj_next_content(oid, NOTHING);
}
     
void
remove_content(Objid where, Objid oid)
{
    Objid lid; 
    if (obj_contents(where) == oid) 
	set_obj_contents(where, obj_next_content(oid));
    else { 
	for (lid = obj_contents(where); lid != NOTHING;
	     lid = obj_next_content(lid)) {
	    if (obj_next_content(lid) == oid) { 
		set_obj_next_content(lid, obj_next_content(oid));
		break; 
	    } 
	} 
    } 
    set_obj_next_content(oid, NOTHING);
}

Var
do_change_parent(Objid thing, Objid parent)
{
    Objid	oid, tid, common, old_parent;
    int		new_props, common_props, old_props;
    Var		r;

    /*
     * check for class hierarchy violations
     */
    for (oid = parent; oid != NOTHING; oid = obj_parent(oid)) {
	if (oid == thing) {
	    r.type = TYPE_ERR;
	    r.v.err = E_RECMOVE;
	    return r;
	}
    }

    /*
     * check for a common ancestor between old & new parent
     */
    common = NOTHING;
    for (oid = thing; oid != NOTHING; oid = obj_parent(oid)) {
	for (tid = parent; tid != NOTHING; tid = obj_parent(tid)) {
	    if (oid == tid) {
		common = oid;
		goto endouter;
	    }
	}
    }
    endouter:

    if (common != NOTHING)
	common_props = count_properties(common);
    else
	common_props = 0;

    old_parent = obj_parent(thing);
    old_props = count_properties(old_parent) - common_props;
    new_props = count_properties(parent) - common_props;

    if (old_parent != NOTHING) 
	remove_child(old_parent, thing);

    if (parent != NOTHING) 
	add_child(parent, thing);

    set_obj_parent(thing, parent);

    fix_props(thing, 0, old_props, new_props, common_props);

    r.type = TYPE_NUM;
    r.v.num = 0;
    return r;
}

/**** objects ****/

Objid
create_object(const char *name, Objid parent, Objid owner)
{
    Object	*o, *p;
    Objid	oid;
    int		i, nprops;

    if (valid(owner)  &&  !decr_quota(owner))
	return NOTHING;

    o = new_object();
    oid = o->id;
    o->name = str_dup(name);
    o->verbdefs = 0;
    o->propdefs.max_length = 0;
    o->propdefs.cur_length = 0;
    o->propdefs.l = 0;
    o->flags = 0;

    /*
     * put new object in class hierarchy
     */
    o->parent = parent;
    o->child = o->sibling = NOTHING;
    if (parent != NOTHING)
	add_child(parent, oid);

    /*
     * transfer ownership to player
     * if owner == NOTHING, make object own itself.
     */
    if (owner == NOTHING)
	o->owner = oid;
    else
	o->owner = owner;

    o->location = o->contents = o->next = NOTHING;
    if ((nprops = count_properties(oid)) != 0) {
	/* If this newly-created object has properties, then it must
	   have a valid parent. */
	p = retrieve(parent);
	o->propval = (Pval *) mymalloc(nprops * sizeof(Pval), M_PVAL);
	for (i = 0; i < nprops; i++) {
	    (o->propval)[i] = (p->propval)[i];
	    (o->propval)[i].var.type = TYPE_CLEAR;
	    if ((o->propval)[i].perms & (1<<PP_CHOWN)) {
		(o->propval)[i].owner = o->owner;
	    }
	}
    } else {
	o->propval = 0;
    }
    return o->id;
}

void
destroy_object(Objid oid)
{
    Object     *obj = retrieve(oid);
    Verbdef    *v, *w;
    int 	i, prop_count;

    incr_quota(obj_owner(oid));
    if (is_player(oid)) {
	Var		t;

	t.type = TYPE_OBJ;
	t.v.obj = oid;
	players_list = setremove(players_list, t);
    }
    prop_count = count_properties(oid);
    for (i = 0; i < prop_count; i++)
	free_var(obj->propval[i].var);
    if (obj->propval)
	myfree((void *) obj->propval, M_PVAL);
    free_str(obj_name(oid));
    for (v = obj->verbdefs; v; v = w) {
	if (v->program)
	    free_program(v->program);
	free_str(v->name);
	w = v->next;
	myfree((void *) v, M_VERBDEF);
    }
    for (i = 0; i < obj->propdefs.cur_length; i++)
	free_str(obj->propdefs.l[i].name);
    if (obj->propdefs.l)
	myfree(obj->propdefs.l, M_PROPDEF);
    free_object(oid);
}

Var
change_parent(Objid oid, Objid parent, Objid progr)
{
    Var		r;

    if (!valid(oid)
	|| (obj_owner(oid) != progr && !is_wizard(progr))
	|| (!valid(parent) && parent != NOTHING)
	|| (valid(parent) && !Obj_perm(progr, parent, FLAG_FERTILE))) {
	r.type = TYPE_ERR;
	r.v.err = E_PERM;
	return r;
    } else if (!check_props(parent, oid)) {
	r.type = TYPE_ERR;
	r.v.err = E_INVARG;
	return r;
    } else
	return do_change_parent(oid, parent);

}

Objid
renumber_object(Objid old)
{
    Objid	new;
    Object     *o;

    for (new = 0; new < old; new++) {
	if (objects[new] == 0) {
	    /* Change the identity of the object. */
	    o = objects[new] = objects[old];
	    objects[old] = 0;
	    objects[new]->id = new;

	    /* Fix up the parent/children hierarchy */
	    {
		Objid  *oidp;
		Object *c;

		if (o->parent != NOTHING) {
		    oidp = &(retrieve(o->parent)->child);
		    while (*oidp != old  &&  *oidp != NOTHING)
			oidp = &(retrieve(*oidp)->sibling);
		    if (*oidp == NOTHING)
			panic("Object not in parent's children list");
		    *oidp = new;
		}

		for (c = retrieve(o->child); c; c = retrieve(c->sibling))
		    c->parent = new;
	    }

	    /* Fix up the contents/location hierarchy */
	    {
		Objid  *oidp;
		Object *c;

		if (o->location != NOTHING) {
		    oidp = &(retrieve(o->location)->contents);
		    while (*oidp != old  &&  *oidp != NOTHING)
			oidp = &(retrieve(*oidp)->next);
		    if (*oidp == NOTHING)
			panic("Object not in location's contents list");
		    *oidp = new;
		}

		for (c = retrieve(o->contents); c; c = retrieve(c->next))
		    c->location = new;
	    }

	    /* Fix up the list of players, if necessary */
	    if (is_player(new)) {
		int	i;

		for (i = 1; i <= players_list.v.list[0].v.num; i++) {
		    if (players_list.v.list[i].v.obj == old) {
			players_list.v.list[i].v.obj = new;
			break;
		    }
		}
	    }

	    /* Fix the owners of verbs, properties and objects */
	    {
		Objid	oid;

		for (oid = 0; oid < nobjects; oid++) {
		    Object     *o = retrieve(oid);
		    Verbdef    *v;
		    Pval       *p;
		    int		i, count = count_properties(oid);

		    if (!o)
			continue;

		    if (o->owner == new)
			o->owner = NOTHING;
		    else if (o->owner == old)
			o->owner = new;

		    for (v = o->verbdefs; v; v = v->next) {
			if (v->owner == new)
			    v->owner = NOTHING;
			else if (v->owner == old)
			    v->owner = new;
		    }

		    p = o->propval;
		    for (i = 0; i < count; i++) {
			if (p[i].owner == new)
			    p[i].owner = NOTHING;
			else if (p[i].owner == old)
			    p[i].owner = new;
		    }
		}
	    }
	    return new;
	}
    }
    return old;
}

/* queries on objects */

int
valid(Objid oid)
{
    return(retrieve(oid) != 0);
}

Objid
obj_owner(Objid oid)
{
    return(retrieve(oid)->owner);
}

const char *
obj_name(Objid oid)
{
    return(retrieve(oid)->name);
}

Objid
obj_location(Objid oid)
{
    return(retrieve(oid)->location);
}

Objid
obj_parent(Objid oid)
{
    return(retrieve(oid)->parent);
}

Objid    
obj_child(Objid oid)
{
    return(retrieve(oid)->child);
}

Objid    
obj_sibling(Objid oid)
{
    return(retrieve(oid)->sibling);
}

Objid    
obj_contents(Objid oid)
{
    return(retrieve(oid)->contents);
}

Objid    
obj_next_content(Objid oid)
{
    return(retrieve(oid)->next);
}

int      
obj_flags(Objid oid)
{
    return(retrieve(oid)->flags);
}

/* modifying objects */

void
set_obj_owner(Objid oid, Objid owner)
{
    retrieve(oid)->owner = owner;
}

void
set_obj_name(Objid oid, const char *name)
{
    retrieve(oid)->name = name;
}

void
set_obj_location(Objid oid, Objid location)
{
    retrieve(oid)->location = location;
}

void
set_obj_parent(Objid oid, Objid parent)
{
    retrieve(oid)->parent = parent;
}

void
set_obj_child(Objid oid, Objid child)
{
    retrieve(oid)->child = child;
}

void
set_obj_sibling(Objid oid, Objid sibling)
{
    retrieve(oid)->sibling = sibling;
}

void
set_obj_contents(Objid oid, Objid contents)
{
    retrieve(oid)->contents = contents;
}

void
set_obj_next_content(Objid oid, Objid next_content)
{
    retrieve(oid)->next = next_content;
}

void     
set_obj_flags(Objid oid, int flags)
{
    retrieve(oid)->flags = flags;
}

char rcsid_db[] = "$Id: db.c,v 1.7 1992/10/23 23:03:47 pavel Exp $";

/* $Log: db.c,v $
 * Revision 1.7  1992/10/23  23:03:47  pavel
 * Added copyright notice.
 *
 * Revision 1.6  1992/10/23  19:24:31  pavel
 * Eliminated all uses of the useless macro NULL..
 *
 * Revision 1.5  1992/10/21  03:02:35  pavel
 * Converted to use new automatic configuration system.
 *
 * Revision 1.4  1992/10/17  20:23:00  pavel
 * Global rename of strdup->str_dup, strref->str_ref, vardup->var_dup, and
 * varref->var_ref.
 * Added a variable initialization to suppress warnings from naive compilers.
 * Made sure to remove old DB file before trying to rename new one into its
 * place, since ANSI doesn't guarantee that that will work otherwise.
 *
 * Revision 1.3  1992/10/06  18:21:11  pavel
 * Fixed up db_flush() to obviously return a value, circumventing a warning
 * from dumb compilers.
 *
 * Revision 1.2  1992/09/15  15:53:23  pjames
 * Renamed FLUSH choices.
 *
 * Revision 1.1  1992/09/14  18:42:30  pjames
 * Initial RCS-controlled version.
 */
