/******************************************************************************
  Copyright (c) 1992 Xerox Corporation.  All rights reserved.
  Portions of this code were written by Stephen White, aka ghond.
  Use and copying of this software and preparation of derivative works based
  upon this software are permitted.  Any distribution of this software or
  derivative works must comply with all applicable United States export
  control laws.  This software is made available AS IS, and Xerox Corporation
  makes no warranty about the software, its performance or its conformity to
  any specification.  Any person obtaining a copy of this software is requested
  to send their name and post office or electronic mail address to:
    Pavel Curtis
    Xerox PARC
    3333 Coyote Hill Rd.
    Palo Alto, CA 94304
    Pavel@Xerox.Com
 *****************************************************************************/

#include "config.h"
#include "db.h"
#include "list.h"
#include "log.h"
#include "storage.h"
#include "unparse.h"
#include "utils.h"

int
count_properties(Objid oid)
{
    Object	*c;
    int		nprop = 0;

    for (c = retrieve(oid); c; c = retrieve(c->parent))
	nprop += c->propdefs.cur_length;
    
    return nprop;
}

void
fix_props(Objid oid, int parent_local, int old, int new, int common)
{
    Object     *me = retrieve(oid);
    Object     *parent = retrieve(me->parent);
    Pval       *new_pvals;
    int		local = parent_local;
    int		i;
    Objid	c;

    local += me->propdefs.cur_length;

    for (i = local; i < local + old; i++)
	free_var(me->propval[i].var);

    if (local + common + new != 0) {
	new_pvals = (Pval *) mymalloc((local + new + common) * sizeof(Pval),
				      M_PVAL);
	for (i = 0; i < local; i++)
	    new_pvals[i] = me->propval[i];
	for (i = 0; i < new; i++) {
	    Pval	pv;

	    pv = parent->propval[parent_local + i];
	    new_pvals[local + i] = pv;
	    new_pvals[local + i].var.type = TYPE_CLEAR;
	    if (pv.perms & (1 << PP_CHOWN))
		new_pvals[local + i].owner = me->owner;
	}
	for (i = 0; i < common; i++)
	    new_pvals[local + new + i] = me->propval[local + old + i];
    } else
	new_pvals = 0;

    if (me->propval)
	myfree((void *) me->propval, M_PVAL);
    me->propval = new_pvals;

    for (c = me->child; c != NOTHING; c = retrieve(c)->sibling)
	fix_props(c, local, old, new, common);
}

Propdef *
match_property(const char *pname, Objid oid, Propid *pid)
{
    Object	*c;
    unsigned int hash = str_hash(pname);
    
    if (match_pseudo_property(pname, pid)) {
	return (Propdef *) 1;
    }

    *pid = 0;
    for (c = retrieve(oid); c; c = retrieve(c->parent)) {
	Proplist *props = &(c->propdefs);
	Propdef  *defs  = props->l;
	unsigned  length = props->cur_length;
	unsigned  loop;

	for (loop = 0; loop < length; loop++) {
	    if (defs[loop].hash_value == hash
		&& !mystrcasecmp(defs[loop].name, pname))
		return &(defs[loop]);
	    (*pid)++;
	}
    }
    
    *pid = FAILED_MATCH;
    return 0;
}

int
check_props(Objid parent, Objid thing)
{
    /* Return true iff PARENT defines no properties that are also defined by */
    /* either THING or any of THING's descendants. */

    Propid	junk;
    Objid	c;
    Proplist   *props = &retrieve(thing)->propdefs;
    unsigned    loop, length = props->cur_length;

    for (loop = 0; loop < length; loop++)
	if (match_property(props->l[loop].name, parent, &junk))
	    return 0;
    
    for (c = retrieve(thing)->child; c != NOTHING; c = retrieve(c)->sibling)
	if (!check_props(parent, c))
	    return 0;

    return 1;
}

Var
get_property_list(Objid oid, Objid progr)
{
    Var r;
    Object *obj = retrieve(oid);
    
    if (!valid(oid)) {
	r.type = TYPE_ERR;
	r.v.err = E_INVARG;
    } else if (!Obj_perm(progr, oid, FLAG_READ)) {
	r.type = TYPE_ERR;
	r.v.err = E_PERM;
    } else {
	Var		prop;
	unsigned	i;/*, count = 0;*/
	
	r = new_list(obj->propdefs.cur_length);
	prop.type = TYPE_STR;
	for (i = 1; i <= obj->propdefs.cur_length; i++) {
	    prop.v.str = str_ref(obj->propdefs.l[i-1].name);
	    r.v.list[i] = prop;
	}
    }
    return r;
}

static void
remove_prop(Objid what, int pos)
{
    Pval	*npval;
    Object	*obj;
    int		i, nprops;

    obj = retrieve(what);
    free_var((obj->propval)[pos].var);	/* free deleted property */

    nprops = count_properties(what);
    if (nprops) {
	npval = (Pval *) mymalloc(nprops * sizeof(Pval), M_PVAL);
	for (i = 0; i < pos; i++)
	    npval[i] = obj->propval[i];
	for (i = pos; i < nprops; i++)
	    npval[i] = obj->propval[i+1];
    } else
	npval = 0;

    if (obj->propval)
	myfree((void *) obj->propval, M_PVAL);
    obj->propval = npval;
}

static void
remove_prop_recursively(Objid root, int root_pos)
{
    Objid	c;

    remove_prop(root, root_pos);
    for (c = obj_child(root); c != NOTHING; c = obj_sibling(c)) {
	int	new_prop_count = retrieve(c)->propdefs.cur_length;
	remove_prop_recursively(c, new_prop_count + root_pos);
    }
}

enum error 
delete_property(Objid oid, const char *pname, Objid progr)
{
    if (!valid(oid))
	return E_INVARG;
    else if (!Obj_perm(progr, oid, FLAG_WRITE))
	return E_PERM;
    else {
	Proplist   *props = &retrieve(oid)->propdefs;
	unsigned int hash = str_hash(pname);
	unsigned    count = props->cur_length;
	unsigned    max   = props->max_length;
	unsigned    loop;
	
	for (loop = 0; loop < count; loop++) {
	    Propdef p;
	    
	    p = props->l[loop];
	    if (p.hash_value == hash && !mystrcasecmp(p.name, pname)) {
		if (max > 8
		    && props->cur_length <= ((max * 3) / 8)) {
		    unsigned new_size = max / 2;
		    unsigned index;
		    Propdef *new_props =
			mymalloc(new_size * sizeof(Propdef), M_PROPDEF);
		    
		    for (index = 0; index < loop; index++)
			new_props[index] = props->l[index];
		    for (index = loop+1; index < count; index++)
			new_props[index - 1] = props->l[index];
		    myfree((void *) props->l, M_PROPDEF);
		    props->l = new_props;
		    props->max_length = new_size;
		} else {
		    unsigned index;
		    for (index = loop+1; index < count; index++)
			props->l[index - 1] = props->l[index];
		}
		--props->cur_length;
		if (p.name)
		    free_str(p.name);
		remove_prop_recursively(oid, loop);
		return E_NONE;
	    }
	}
	return E_PROPNF;
	
    }
}
     
void
get_bi_prop(Objid oid, int pid, Var *ret)
{
    switch (pid) {
      case LOCATION:
	ret->type = TYPE_OBJ;
	ret->v.obj = obj_location(oid);
	break;
      case OWNER:
	ret->type = TYPE_OBJ;
	ret->v.obj = obj_owner(oid);
	break;
      case NAME:
	ret->type = TYPE_STR;
	ret->v.str = str_ref(obj_name(oid));
	break;
      case CONTENTS:
	{
	    Var		item, r;
	    Objid	c;
	    int		i, count = 0;
	    
	    for (c = obj_contents(oid);
		 c != NOTHING;
		 c = obj_next_content(c))
		count++;
	    r = new_list(count);
	    item.type = TYPE_OBJ;
	    for (c = obj_contents(oid), i = 1;
		 c != NOTHING;
		 c = obj_next_content(c), i++) {
		item.v.obj = c;
		r.v.list[i] = item;
	    }
	    *ret = r;
	}
	break;
      case FLAG - FLAG_PLAYER:
	ret->type = TYPE_NUM;
	ret->v.num = is_player(oid);
	break;
      case FLAG - FLAG_PROGRAMMER:
	ret->type = TYPE_NUM;
	ret->v.num = is_programmer(oid);
	break;
      case FLAG - FLAG_WIZARD:
	ret->type = TYPE_NUM;
	ret->v.num = is_wizard(oid);
	break;
      case FLAG - FLAG_READ:
	ret->type = TYPE_NUM;
	ret->v.num = ((obj_flags(oid) & (1 << FLAG_READ)) != 0);
	break;
      case FLAG - FLAG_WRITE:
	ret->type = TYPE_NUM;
	ret->v.num = ((obj_flags(oid) & (1 << FLAG_WRITE)) != 0);
	break;
      case FLAG - FLAG_FERTILE:
	ret->type = TYPE_NUM;
	ret->v.num = ((obj_flags(oid) & (1 << FLAG_FERTILE)) != 0);
	break;
      default:
	errlogf("GET_BI_PROP: Unknown PID: %d\n", pid);
	ret->type = TYPE_NUM;
	ret->v.num = 0;
	break;
    }
}

enum error
set_bi_prop(Objid oid, int pid, Var value, Objid progr)
{
    if (progr != obj_owner(oid) && !is_wizard(progr))
	return E_PERM;
    
    switch (pid) {
      case NAME:
	if (value.type != TYPE_STR)
	    return E_TYPE;
	if (is_player(oid) && !is_wizard(progr))
	    return E_PERM;
	if (obj_name(oid))
	    free_str(obj_name(oid));
	set_obj_name(oid, value.v.str);
	break;
      case OWNER:
	if (value.type != TYPE_OBJ)
	    return E_TYPE;
	if (!is_wizard(progr))
	    return E_PERM;
	set_obj_owner(oid, value.v.obj);
	break;
      case FLAG - FLAG_PROGRAMMER:
      case FLAG - FLAG_WIZARD:
	if (!is_wizard(progr))
	    return E_PERM;    
	if (!is_player(oid))
	    return E_INVARG;
	if (pid == FLAG - FLAG_WIZARD  &&  is_true(value)) {
	    logf("WIZARDED: #%d by programmer #%d\n", oid, progr);
	    print_error_backtrace("Wizard bit set.", 1);
	}
	/* fall through to ... */
      case FLAG - FLAG_READ:
      case FLAG - FLAG_WRITE:
      case FLAG - FLAG_FERTILE:
	{
	    int	flag = FLAG - pid;
	    
	    set_obj_flags(oid, obj_flags(oid) & ~(1 << flag));
	    if (is_true(value))
		set_obj_flags(oid, obj_flags(oid) | (1 << flag));
	}
	break;
      default:
	return E_PERM;
    }
    
    return E_NONE;
}

static enum error
get_prop(Objid oid, const char *name, Var *value,
	 Objid progr,
	 int keep_clear,	/* walk up parents if TYPE_CLEAR */
	 int safe,		/* don't check perms */
	 int check_bi)		/* look at built in functions */
{
    int		pid;
    Objid       o = oid;

    if (!valid(oid))
	return E_INVARG;
    
    if (!match_property(name, oid, &pid))
	return E_PROPNF;

    if (pid < 0) {
	if (!check_bi)
	    return E_PROPNF;
	get_bi_prop(oid, pid, value);
	return E_NONE;
    } else {
	Pval	*prop = retrieve(oid)->propval + pid;
	
	if (safe || Prop_perm(progr, prop, PP_READ)) {
	    while (!keep_clear && prop->var.type == TYPE_CLEAR) {
		pid -= retrieve(o)->propdefs.cur_length;
		o = obj_parent(o);
		prop = retrieve(o)->propval + pid;
	    }
	    *value = var_ref(prop->var);
	    return E_NONE;
	} else {
	    return E_PERM;
	}
    }
}
     
enum error
get_non_bi_prop(Objid oid, const char *name, Var *value, int keep_clear)
{ 
    return get_prop(oid, name, value, oid /* dummy */, keep_clear, 1, 0);
}

enum error
get_prop_safe(Objid oid, const char *name, Var *value, int keep_clear)
{
    return get_prop(oid, name, value, oid /* dummy */, keep_clear, 1, 1);
}

enum error
get_prop_val(Objid oid, const char *name, Var *value,
	     Objid progr, int keep_clear)
{
    return get_prop(oid, name, value, progr, keep_clear, 0, 1);
}

enum error
set_prop(Objid oid, const char *pname, Var value, Objid progr,
	 int safe)		/* don't check permissions if safe = 1 */
{
    int		pid;

    if (!valid(oid))
	return E_INVARG;
    
    if (!match_property(pname, oid, &pid))
	return E_PROPNF;
    
    if (pid < 0) {
	if (!safe && progr != obj_owner(oid) && !is_wizard(progr)) {
	    return E_PERM;
	} else {
	    return set_bi_prop(oid, pid, var_ref(value), progr);
	}
    } else {
	Pval   *prop = retrieve(oid)->propval + pid;
	int     local = (pid < retrieve(oid)->propdefs.cur_length);

	if (value.type == TYPE_CLEAR && local) {
	    return E_INVARG;
	} else if (safe || Prop_perm(progr, prop, PP_WRITE)) {
	    free_var(prop->var);
	    prop->var = var_ref(value);
	} else {
	    return E_PERM;
	}
    }
    return E_NONE;
}

enum error
set_prop_val(Objid oid, const char *pname, Var value, Objid progr)
{
    return set_prop(oid, pname, value, progr, 0);
}

enum error
set_prop_safe(Objid oid, const char *pname, Var value)
{
    return set_prop(oid, pname, value, oid /*dummy*/, 1);
}

static int
property_defined_at_or_below(const char *pname, Objid oid)
{
    /* Return true iff some descendant of OID defines a property named PNAME */
    
    Objid	c;
    unsigned int hash = str_hash(pname);
    Proplist   *props = &retrieve(oid)->propdefs;
    unsigned    length = props->cur_length;
    unsigned    loop;
    
    for (loop = 0; loop < length; loop++)
	if (props->l[loop].hash_value == hash
	    && !mystrcasecmp(props->l[loop].name, pname))
	    return 1;
    
    for (c = obj_child(oid); c != NOTHING; c = obj_sibling(c))
	if (property_defined_at_or_below(pname, c))
	    return 1;
    
    return 0;
}

static void
insert_prop(Objid what, int pos, Pval pv)
{
    Pval	*npval;
    Object	*o;
    int		i, nprops;

    o = retrieve(what);

    /*
     * allocate new property area
     */
    nprops = count_properties(what);
    npval = (Pval *) mymalloc(nprops * sizeof(Pval), M_PVAL);
    for (i = 0; i < pos; i++)
	npval[i] = o->propval[i];
    for (i = pos; i < nprops - 1; i++)
	npval[i+1] = o->propval[i];
    npval[pos] = pv;	/* assign new property */
    npval[pos].var = var_ref(pv.var);
    if (npval[pos].perms & (1<<PP_CHOWN))
	npval[pos].owner = o->owner;

    if (o->propval)
	myfree((void *) o->propval, M_PVAL);
    o->propval = npval;
}

static void
insert_prop_recursively(Objid root, int root_pos, Pval pv)
{
    Objid	c;

    insert_prop(root, root_pos, pv);
    pv.var.type = TYPE_CLEAR;  /* do after initial insert_prop so only
				   children will be TYPE_CLEAR */
    for (c = obj_child(root); c != NOTHING; c = obj_sibling(c)) {
	int	new_prop_count = retrieve(c)->propdefs.cur_length;
	insert_prop_recursively(c, new_prop_count + root_pos, pv);
    }
}

unsigned int
str_hash(const char *s)
{
    unsigned int ans = 0;
    int loop, len = strlen(s), offset = 0;

    for (loop = 0; loop < len; loop++) {
	ans = ans ^ (cmap[(unsigned char) s[loop]] << offset++);
	if (offset == 25)
	    offset = 0;
    }
    return ans;
}

Propdef
new_propdef(const char *name)
{
    Propdef	newprop;

    newprop.name = str_ref(name);
    newprop.hash_value = str_hash(name);
    return newprop;
}

static int
parse_prop_perms(const char *perm_str)
{
    int	new_perms = 0;
    
    while (*perm_str) {
	switch (*(perm_str++)) {
	  case 'r':
	  case 'R':
	    new_perms |= (1 << PP_READ);
	    break;
	  case 'w':
	  case 'W':
	    new_perms |= (1 << PP_WRITE);
	    break;
	  case 'c':
	  case 'C':
	    new_perms |= (1 << PP_CHOWN);
	    break;
	  default:
	    return -1;
	}
    }
    
    return new_perms;
}

static int
is_prop_info(Var *list)
{
    return (list[0].v.num == 2
	    && list[1].type == TYPE_OBJ
	    && list[2].type == TYPE_STR);
}

/*
 * Add a new property to the given object (and all of its descendants).
 * Assume that permissions have already been checked, that the property name
 * is legal and that no property with this name already exists on the object.
 * Return the index of the new property.
 */
enum error
add_property(Objid oid, const char *pname, Objid progr, Var init, Var info)
{
    Object     *obj = retrieve(oid);
    Pval	pv;
    unsigned    loop;
    int 	pid;
    Objid       owner;
    int         perms;

    if (!is_prop_info(info.v.list))
	return E_TYPE;
    owner = info.v.list[1].v.obj;
    perms = parse_prop_perms(info.v.list[2].v.str);

    if (!valid(oid)
	|| match_property(pname, oid, &pid)
	|| property_defined_at_or_below(pname, oid)
	|| perms < 0
	|| !valid(owner)) 
	return E_INVARG;
    else if (!Obj_perm(progr, oid, FLAG_WRITE)
	     || (!is_wizard(progr) && (progr != owner))) 
	return E_PERM;
    else {
	
	if (obj->propdefs.cur_length == obj->propdefs.max_length) {
	    Propdef    *old_props = obj->propdefs.l;	
	    unsigned new_size = (obj->propdefs.max_length == 0) ? 8
		: 2 * obj->propdefs.max_length;
	    
	    obj->propdefs.l = mymalloc(new_size * sizeof(Propdef), M_PROPDEF);
	    for (loop = 0; loop < obj->propdefs.max_length; loop++)
		obj->propdefs.l[loop] = old_props[loop];
	    if (old_props)
		myfree(old_props, M_PROPDEF);
	    obj->propdefs.max_length = new_size;
	}    
	obj->propdefs.l[obj->propdefs.cur_length++] = new_propdef(pname);
	
	pv.var = init;
	pv.owner = owner;
	pv.perms = perms;
	
	insert_prop_recursively(oid, obj->propdefs.cur_length - 1, pv);
    }
    return E_NONE;
}

Var 
get_prop_info(Objid oid, const char *pname, Objid progr)
{
    Var 	r;
    int 	pid;
    Object     *obj = retrieve(oid);

    if (!valid(oid)) {
	r.type = TYPE_ERR;
	r.v.err = E_INVARG;
    } else if (!match_property(pname, oid, &pid) || pid < 0) {
	r.type = TYPE_ERR;
	r.v.err = E_PROPNF;
    } else {
	Pval   *pval = obj->propval + pid;
	int	i = 0;
	char   *s;

	if (!Prop_perm(progr, pval, PP_READ)) {
	    r.type = TYPE_ERR;
	    r.v.err = E_PERM;
	} else {
	    r = new_list(2);
	    r.v.list[1].type = TYPE_OBJ;
	    r.v.list[1].v.obj = pval->owner;
	    r.v.list[2].type = TYPE_STR;
	    r.v.list[2].v.str = s = str_dup("xxx");
	    if (pval->perms & (1 << PP_READ)) s[i++] = 'r';
	    if (pval->perms & (1 << PP_WRITE)) s[i++] = 'w';
	    if (pval->perms & (1 << PP_CHOWN)) s[i++] = 'c';
	    s[i] = '\0';
	}
    }
    return r;
}

enum error 
set_prop_info(Objid oid, const char *pname, Objid progr, Var info)
{
    if (!valid(oid))
	return E_INVARG;
    if (!is_prop_info(info.v.list)) 
	return E_TYPE;
    else {
	Objid		new_owner = info.v.list[1].v.obj;
	const char *	perm_str = info.v.list[2].v.str;
	int		new_perms = parse_prop_perms(perm_str);
	int		pid;

	if (!match_property(pname, oid, &pid) || pid < 0)
	    return E_PROPNF;
	else {
	    Pval       *pval = retrieve(oid)->propval + pid;
	    
	    if ((!is_wizard(progr) && new_owner != pval->owner)
		|| !Prop_perm(progr, pval, PP_WRITE)) 
		return E_PERM;
	    else if (new_perms < 0 || !valid(new_owner)) 
		return E_INVARG;
	    else {
		pval->owner = new_owner;
		pval->perms = new_perms;
		return E_NONE;
	    }
	}
    }
}


char rcsid_db_property[] = "$Id: db_property.c,v 1.4 1992/10/23 23:03:47 pavel Exp $";

/* $Log: db_property.c,v $
 * Revision 1.4  1992/10/23  23:03:47  pavel
 * Added copyright notice.
 *
 * Revision 1.3  1992/10/21  03:02:35  pavel
 * Converted to use new automatic configuration system.
 *
 * Revision 1.2  1992/10/17  20:23:38  pavel
 * Global rename of strdup->str_dup, strref->str_ref, vardup->var_dup, and
 * varref->var_ref.
 *
 * Revision 1.1  1992/09/14  18:42:30  pjames
 * Initial RCS-controlled version.
 *
 */
