/******************************************************************************
  Copyright (c) 1992 Xerox Corporation.  All rights reserved.
  Portions of this code were written by Stephen White, aka ghond.
  Use and copying of this software and preparation of derivative works based
  upon this software are permitted.  Any distribution of this software or
  derivative works must comply with all applicable United States export
  control laws.  This software is made available AS IS, and Xerox Corporation
  makes no warranty about the software, its performance or its conformity to
  any specification.  Any person obtaining a copy of this software is requested
  to send their name and post office or electronic mail address to:
    Pavel Curtis
    Xerox PARC
    3333 Coyote Hill Rd.
    Palo Alto, CA 94304
    Pavel@Xerox.Com
 *****************************************************************************/

#include "my-stdlib.h"

#include "config.h"
#include "db.h"
#include "execute.h"
#include "list.h"
#include "match.h"
#include "parse_cmd.h"
#include "parser.h"
#include "storage.h"
#include "unparse.h"
#include "utils.h"

/**** verbs ****/

static Verbdef *
match_verbdef(const char *vname, Objid oid, int  *vid)
{
    Verbdef    *v;
    int		num;
    char       *p;

    num = strtol(vname, &p, 10);
    if (*vname == '\0' || *p != '\0')
	num = -1;

    *vid = 0;
    v = retrieve(oid)->verbdefs;
    for (; v; v = v->next)  {
	if (verbcasecmp(v->name, vname) || *vid == num)
	    return v;
	(*vid)++;
    }
    *vid = FAILED_MATCH;
    return 0;
}

Verbdef *
find_nth_verbdef(Objid oid, unsigned n)
{
    Verbdef    *v;
    unsigned	i;

    for (i = 1, v = retrieve(oid)->verbdefs; v; i++, v = v->next)
	if (i == n)
	    return v;

    return 0;
}

Var 
get_verb_list(Objid oid, Objid progr)
{
    Var r;
    
    if (!valid(oid)) {
	r.type = TYPE_ERR;
	r.v.err = E_INVARG;
    } else if (!Obj_perm(progr, oid, FLAG_READ)) {
	r.type = TYPE_ERR;
	r.v.err = E_PERM;
    } else {
	Verbdef    *v;
	Var	    verb;
	int	    i, count = 0;
	
	for (v = retrieve(oid)->verbdefs; v; v = v->next)
	    count++;
	r = new_list(count);
	verb.type = TYPE_STR;
	for (v = retrieve(oid)->verbdefs, i = 1; v; v = v->next, i++) {
	    verb.v.str = str_ref(v->name);
	    r.v.list[i] = verb;
	}
    }
    return r;
}

static void
lister(void *data, const char *line)
{
    Var	*r = (Var *) data;
    Var	v;
    
    v.type = TYPE_STR;
    v.v.str = str_dup(line);
    *r = listappend(*r, v);
}

static int
is_verb_info(Var *list)
{
    return (list[0].v.num == 3
	    && list[1].type == TYPE_OBJ
	    && list[2].type == TYPE_STR
	    && list[3].type == TYPE_STR);
}

static int
is_args(Var *list)
{
    return (list[0].v.num == 3
	    && list[1].type == TYPE_STR
	    && list[2].type == TYPE_STR
	    && list[3].type == TYPE_STR);
}

static int
is_code(Var *list)
{
    int		i;
    
    for (i = 1; i <= list[0].v.num; i++)
	if (list[i].type != TYPE_STR)
	    return 0;
    return 1;
}

static int
parse_verb_perms(const char *perm_str)
{
    int	new_perms = 0;
    
    while (*perm_str) {
	switch (*(perm_str++)) {
	  case 'r':
	  case 'R':
	    new_perms |= (1 << VP_READ);
	    break;
	  case 'w':
	  case 'W':
	    new_perms |= (1 << VP_WRITE);
	    break;
	  case 'x':
	  case 'X':
	    new_perms |= (1 << VP_EXEC);
	    break;
	  case 'd':
	  case 'D':
	    new_perms |= (1 << VP_DEBUG);
	    break;
	  default:
	    return -1;
	}
    }
    
    return new_perms;
}

struct parser_state {
    Var		code;		/* a list of strings */
    int		cur_string;	/* which string? */
    int		cur_char;	/* which character in that string? */
    Var		errors;		/* a list of strings */
};

static enum vopt
match_option(const char *opt)
{
    if (!mystrcasecmp(opt, "none")) return OPT_NONE;
    if (!mystrcasecmp(opt, "any")) return OPT_ANY;
    if (!mystrcasecmp(opt, "this")) return OPT_THIS;
    return OPT_FAIL;
}

enum error
add_verb(Objid oid, Objid progr, Var info, Var args)
{
    if (!is_verb_info(info.v.list) || !is_args(args.v.list))
	return E_TYPE;
    else if (!valid(oid))
	return E_INVARG;
    else {
	Objid		owner = info.v.list[1].v.obj;
	int		perms = parse_verb_perms(info.v.list[2].v.str);
	const char     *names = info.v.list[3].v.str;
	int		dobj = match_option(args.v.list[1].v.str);
	int		prep = match_prep(args.v.list[2].v.str);
	int		iobj = match_option(args.v.list[3].v.str);

	
	while (*names == ' ')
	    names++;
	if (perms < 0
	    || !valid(owner)
	    || *names == '\0'
	    || dobj == OPT_FAIL
	    || prep == FAILED_MATCH
	    || iobj == OPT_FAIL) 
	    return E_INVARG;
	else if (!Obj_perm(progr, oid, FLAG_WRITE)
		   || (!is_wizard(progr) && progr != owner)) 
	    return E_PERM;
	else {
	    Verbdef    *v, *newv;
	    Object     *o = retrieve(oid);
	    
	    newv = (Verbdef *) mymalloc(sizeof(Verbdef), M_VERBDEF);
	    newv->name = str_dup(names);
	    newv->owner = owner;
	    newv->perms = perms | (dobj << VP_DOBJSHIFT)
		| (iobj << VP_IOBJSHIFT);
	    newv->prep = prep;
	    newv->next = 0;
	    newv->program = 0;
	    if (o->verbdefs) {
		for (v = o->verbdefs; v->next; v = v->next)
		    ;
		v->next = newv;
	    } else
		o->verbdefs = newv;
	}
    }
    return E_NONE;
}

enum error
delete_verb(Objid oid, const char *vname, Objid progr)
{
    if (!valid(oid))
	return E_INVARG;
    else if (!Obj_perm(progr, oid, FLAG_WRITE))
	return E_PERM;
    else {
	int		vid;
	Object     *obj = retrieve(oid);
	Verbdef    *v = match_verbdef(vname, oid, &vid);
	
	if (!v) 
	    return E_VERBNF;
	else {
	    Verbdef    *vv;
	    
	    vv = obj->verbdefs;
	    if (vv == v)
		obj->verbdefs = v->next;
	    else {
		while (vv->next != v)
		    vv = vv->next;
		vv->next = v->next;
	    }
	    if (v->program)
		free_program(v->program);
	    if (v->name)
		free_str(v->name);
	    myfree((void *) v, M_VERBDEF);
	}
    }
    return E_NONE;
}

enum error 
get_verb_code(Objid oid, const char *vname, Objid progr, Var *code,
			 int parens, int indent)
{
    if (!valid(oid))
	return E_INVARG;
    else {
	int		vid;
	Verbdef	       *v = match_verbdef(vname, oid, &vid);

	if (!v)
	    return E_VERBNF;
	else if (!Verb_perm(progr, v, VP_READ))
	    return E_PERM;
	else {
	    *code = new_list(0);
	    if (v->program)
		list_prog(v->program, lister, (void *) code, parens, indent,
			  MAIN_VECTOR);
	}
    }
    return E_NONE;
}

Var
set_verb_code(Objid oid, const char *vname, Objid progr, Var code)
{
    Var r;

    if (!valid(oid)) {
	r.type = TYPE_ERR;
	r.v.err = E_INVARG;
    } else if (!is_code(code.v.list)) {
	r.type = TYPE_ERR;
	r.v.err = E_TYPE;
    } else if (!is_programmer(progr)) {
	r.type = TYPE_ERR;
	r.v.err = E_PERM;
    } else {
	int 		vid;
	Verbdef	       *v = match_verbdef(vname, oid, &vid);

	if (!v) {
	    r.type = TYPE_ERR;
	    r.v.err = E_VERBNF;
	} else if (!Verb_perm(progr, v, VP_WRITE)) {
	    r.type = TYPE_ERR;
	    r.v.err = E_PERM;
	} else {
	    Program    *program = parse_list_as_program(code, &r);
	    
	    if (program) {
		if (task_timed_out)
		    free_program(program);
		else {
		    if (v->program)
			free_program(v->program);
		    v->program = program;
		}
	    }
	}
    }
    return r;
}

Var 
get_verb_info(Objid oid, const char *verb_name, Objid progr)
{
    Var r;
    
    if (!valid(oid)) {
	r.type = TYPE_ERR;
	r.v.err = E_INVARG;
    } else {
	int		vid, i = 0;
	Verbdef    *v = match_verbdef(verb_name, oid, &vid);
	char       *s;
	
	if (!v) {
	    r.type = TYPE_ERR;
	    r.v.err = E_VERBNF;
	} else if (!Verb_perm(progr, v, VP_READ)) { 
	    r.type = TYPE_ERR;
	    r.v.err = E_PERM;
	} else {
	    r = new_list(3);
	    r.v.list[1].type = TYPE_OBJ;
	    r.v.list[1].v.obj = v->owner;
	    r.v.list[2].type = TYPE_STR;
	    r.v.list[2].v.str = s = str_dup("xxxx");
	    if (v->perms & (1 << VP_READ)) s[i++] = 'r';
	    if (v->perms & (1 << VP_WRITE)) s[i++] = 'w';
	    if (v->perms & (1 << VP_EXEC)) s[i++] = 'x';
	    if (v->perms & (1 << VP_DEBUG)) s[i++] = 'd';
	    s[i] = '\0';
	    r.v.list[3].type = TYPE_STR;
	    r.v.list[3].v.str = str_ref(v->name);
	}
    }
    return r;
}

enum error 
set_verb_info(Objid oid, const char *vname, Objid progr, Var info)
{
    if (!is_verb_info(info.v.list))
	return E_TYPE;
    else if (!valid(oid))
	return E_INVARG;
    else {
	int		vid;
	Verbdef        *v = match_verbdef(vname, oid, &vid);

	if (!v)
	    return E_VERBNF;
	else {
	    Objid	new_owner = info.v.list[1].v.obj;
	    const char *perm_str = info.v.list[2].v.str;
	    int		new_perms = parse_verb_perms(perm_str);
	    const char *new_name = info.v.list[3].v.str;

	    while (*new_name == ' ')
		new_name++;
	    if ((!is_wizard(progr) && new_owner != v->owner)
		|| !Verb_perm(progr, v, VP_WRITE)) 
		return E_PERM;
	    else if (new_perms < 0 || !valid(new_owner) || *new_name == '\0')
		return E_INVARG;
	    else {
		v->owner = new_owner;
		v->perms &= ~VP_PERMMASK;
		v->perms |= new_perms;
		v->name = str_dup(new_name);
	    }
	}
    }
    return E_NONE;
}

Var 
get_verb_args(Objid oid, const char *vname, Objid progr)
{
    Var r;
    
    if (!valid(oid)) {
	r.type = TYPE_ERR;
	r.v.err = E_INVARG;
    } else {
	int		vid;
	Verbdef        *v = match_verbdef(vname, oid, &vid);
	
	if (!v) {
	    r.type = TYPE_ERR;
	    r.v.err = E_VERBNF;
	} else if (!Verb_perm(progr, v, VP_READ)) {
	    r.type = TYPE_ERR;
	    r.v.err = E_PERM;
	} else {
	    r = new_list(3);
	    r.v.list[1].type = TYPE_STR;
	    r.v.list[1].v.str =
		str_dup(unparse_option((v->perms >> VP_DOBJSHIFT)
				      & VP_OBJMASK));
	    r.v.list[2].type = TYPE_STR;
	    r.v.list[2].v.str = str_dup(unparse_prep(v->prep));
	    r.v.list[3].type = TYPE_STR;
	    r.v.list[3].v.str =
		str_dup(unparse_option((v->perms >> VP_IOBJSHIFT)
				      & VP_OBJMASK));
	}
    }
    return r;
}

enum error 
set_verb_args(Objid oid, const char *vname, Objid progr, Var args)
{
    if (!is_args(args.v.list))
	return E_TYPE;
    else if (!valid(oid))
	return E_INVARG;
    else {
	int		vid;
	Verbdef	       *v = match_verbdef(vname, oid, &vid);
	int		dobj = match_option(args.v.list[1].v.str);
	int		prep = match_prep(args.v.list[2].v.str);
	int		iobj = match_option(args.v.list[3].v.str);

	if (!v) 
	    return E_VERBNF;
	else if (dobj == OPT_FAIL
		 || prep == FAILED_MATCH
		 || iobj == OPT_FAIL) 
	    return E_INVARG;
	else if (!Verb_perm(progr, v, VP_WRITE)) 
	    return E_PERM;
	else {
	    v->perms = ((v->perms & VP_PERMMASK)
			| (dobj << VP_DOBJSHIFT)
			| (iobj << VP_IOBJSHIFT));
	    v->prep = prep;
	}
    }
    return E_NONE;
}

#define M_OPT(tplate, arg, this) ( \
		   (tplate) == (int) OPT_ANY \
		|| ((tplate) == (int) OPT_NONE && (arg) == NOTHING) \
		|| ((tplate) == (int) OPT_THIS && (arg) == (this)) \
		)

Program *
get_program_and_pi(Objid oid, const char *vname, Parse_Info *pi,
		   int match_options)
{
    Verbdef    *v;

    pi->vloc = oid;
    for (v = retrieve(oid)->verbdefs; v; v = v->next) {
	if (verbcasecmp(v->name, vname)) {
	    if (match_options
		   ? (M_OPT((v->perms >> VP_DOBJSHIFT) & VP_OBJMASK,
			    pi->dobj, pi->this)
		      && M_OPT((v->perms >> VP_IOBJSHIFT) & VP_OBJMASK,
			       pi->iobj, pi->this)
		      && (v->prep == AMBIGUOUS || v->prep == pi->prep))
		   : v->perms & (1 << VP_EXEC)) {
		pi->debug = v->perms & (1 << VP_DEBUG);
		pi->progr = v->owner;
		pi->verbname = str_ref(v->name);
		return v->program;
	    }
	}
    }
    return 0;    
}

Verbdef *
find_verbdef(Objid oid, const char *verbname)
{
    Verbdef    *v;

    for (v = retrieve(oid)->verbdefs; v; v = v->next)
	if (verbcasecmp(v->name, verbname))
	    return v;
    return 0;
}

static Verbdef *
check_verbdefs_simple(Objid oid, const char *verbname)
{
    Verbdef    *v;

    for (v = retrieve(oid)->verbdefs; v; v = v->next)
	if (verbcasecmp(v->name, verbname))
	    if (v->perms & (1 << VP_EXEC))
		return v;
    return 0;
}

verbdef_info 
get_verbdef_info(Objid oid, const char *vname)
{
    Objid 		p;
    struct verbdef_info ret;

    ret.ptr = 0;
    if (valid(oid))
	for (p = oid; p != NOTHING; p = obj_parent(p)) {
	    Verbdef *verbdef = check_verbdefs_simple(p, vname);

	    if (verbdef) {
		ret.vloc = p;
		ret.ptr = verbdef;
		return ret;
	    }
	}
    return ret;
}

verbdef_info 
verbdef_info_from_num(Objid oid, int vnum)
{
    verbdef_info	ans;
    Verbdef	       *v = retrieve(oid)->verbdefs;
    
    while (v  &&  vnum > 0) {
	vnum--;
	v = v->next;
    }
    ans.vloc = oid;
    ans.ptr = (void *)v;
    return ans;
}

int
verb_exists(verbdef_info v_i)
{
    return v_i.ptr != 0;
}

Program *
program_from_info(verbdef_info v_i)
{
    return ((Verbdef *)v_i.ptr)->program;
}

void
set_program_in_info(verbdef_info v_i, Program *p)
{
    ((Verbdef *)v_i.ptr)->program = p;
}

const char *
name_from_info(verbdef_info v_i)
{
    return ((Verbdef *)v_i.ptr)->name;
}

Objid
owner_from_info(verbdef_info v_i)
{
    return ((Verbdef *)v_i.ptr)->owner;
}

int
perms_from_info(verbdef_info v_i)
{
    return ((Verbdef *)v_i.ptr)->perms;
}

verbref_info
get_verbref_info(Objid player, const char *vref)
{
    verbref_info ans;
    char       *w = strchr(vref, ':');
    const char *vname;
    Objid	oid;

    if (w == 0  ||  w[1] == '\0') {
	ans.oid = 0;
	ans.verbnum = NOTHING;
	ans.ptr = 0;
	return ans;
    }
    *w = '\0';
    vname = w + 1;
    if (vref[0] == '$')
	oid = get_system_object(vref + 1);
    else
	oid = match_object(player, vref);
    if (valid(oid)) {
	ans.oid = oid;
	ans.ptr = (void *)match_verbdef(vname, oid, &ans.verbnum);
	return ans;
    }
    ans.oid = FAILED_MATCH;
    ans.verbnum = 0;
    ans.ptr = 0;
    return ans;
}

int
check_verbref_perms(verbref_info vr_i, Objid player)
{
    return Verb_perm(player, ((Verbdef *)vr_i.ptr), VP_WRITE);
}

const char *
verbref_name(verbref_info vr_i)
{
    return ((Verbdef *)vr_i.ptr)->name;
}

char rcsid_db_verb[] = "$Id: db_verb.c,v 1.6 1992/10/23 23:03:47 pavel Exp $";

/* $Log: db_verb.c,v $
 * Revision 1.6  1992/10/23  23:03:47  pavel
 * Added copyright notice.
 *
 * Revision 1.5  1992/10/23  19:24:46  pavel
 * Eliminated all uses of the useless macro NULL..
 *
 * Revision 1.4  1992/10/21  03:02:35  pavel
 * Converted to use new automatic configuration system.
 *
 * Revision 1.3  1992/10/17  20:23:58  pavel
 * Global rename of strdup->str_dup, strref->str_ref, vardup->var_dup, and
 * varref->var_ref.
 *
 * Revision 1.2  1992/10/06  18:26:17  pavel
 * Changed name of global Parser_Client to avoid a name clash.
 *
 * Revision 1.1  1992/09/14  18:42:30  pjames
 * Initial RCS-controlled version.
 */
