/******************************************************************************
  Copyright (c) 1992 Xerox Corporation.  All rights reserved.
  Portions of this code were written by Stephen White, aka ghond.
  Use and copying of this software and preparation of derivative works based
  upon this software are permitted.  Any distribution of this software or
  derivative works must comply with all applicable United States export
  control laws.  This software is made available AS IS, and Xerox Corporation
  makes no warranty about the software, its performance or its conformity to
  any specification.  Any person obtaining a copy of this software is requested
  to send their name and post office or electronic mail address to:
    Pavel Curtis
    Xerox PARC
    3333 Coyote Hill Rd.
    Palo Alto, CA 94304
    Pavel@Xerox.Com
 *****************************************************************************/

#include "config.h"
#include "eval_vm.h"
#include "execute.h"
#include "log.h"
#include "storage.h"
#include "structures.h"
#include "tasks.h"
#include "unparse.h"

/**** external functions ****/

vm
new_vm(int stack_size)
{
    vm	the_vm = mymalloc(sizeof(vmstruct), M_VM);

    the_vm->activ_stack = mymalloc(sizeof(activation) * stack_size, M_VM);

    return the_vm;
}

void
free_vm(vm the_vm, int stack_too)
{
    int i;

    if (stack_too) 
	for (i = the_vm->top_activ_stack; i >= 0; i--)
	    free_activation(the_vm->activ_stack[i], 1);
    myfree(the_vm->activ_stack, M_VM);
    myfree(the_vm, M_VM);
}

enum outcome
resume_suspended_task(vm the_vm, Var *result)
{
    Var	v;

    v.type = TYPE_NUM;
    v.v.num = 0;
    set_task_kind(TASK_SUSPENDED);
    return resume_from_previous_vm(the_vm, v, result);
}

enum outcome
resume_reading_vm(vm the_vm, const char *input, Var *result)
{
    Var v;

    if (input) {
	v.type = TYPE_STR;
	v.v.str = input;
    } else {
	v.type = TYPE_ERR;
	v.v.err = E_INVARG;
    }
    set_task_kind(TASK_INPUT);
    return resume_from_previous_vm(the_vm, v, result);
}

activation
top_activ(vm the_vm)
{
    return the_vm->activ_stack[the_vm->top_activ_stack];
}

Objid
progr_of_cur_verb(vm the_vm) 
{
    return top_activ(the_vm).progr;
}

unsigned
suspended_lineno_of_vm(vm the_vm)
{
    activation top;

    top = top_activ(the_vm);
    return get_lineno(top.prog, top.error_pc,
		      (the_vm->top_activ_stack == 0) 
		      ? the_vm->root_activ_vector 
		      : MAIN_VECTOR);
}

/**** read/write data base ****/

void 
write_vm(vm the_vm)
{
    unsigned i;

    db_printf("%u %d %u %u\n",
	      the_vm->top_activ_stack, the_vm->root_activ_vector,
	      the_vm->func_id, the_vm->max_stack_size);
    
    for (i = 0; i <= the_vm->top_activ_stack; i++)
	write_activ(the_vm->activ_stack[i]);
}

int
read_vm(FILE *f, vm *addr_for_the_vm)
{
    unsigned 	i, j, top, func_id, max;
    int		vector;
    char	c;
    vm 		the_vm;

    if (fscanf(f, "%u %d %u", &top, &vector, &func_id) != 3
	|| ((c = fgetc(f)) == ' '
	    ? fscanf(f, "%u", &max) != 1 || fgetc(f) != '\n'
	    : (max = DEFAULT_MAX_STACK_DEPTH, c != '\n'))) {
	errlogf("READ_VM: Bad vm header\n");
	return -1;
    }
    the_vm = new_vm(top + 1);
    the_vm->max_stack_size = max;
    the_vm->top_activ_stack = top;
    the_vm->root_activ_vector = vector;
    the_vm->func_id = func_id;

    for (i = 0; i <= top; i++)
	if (read_activ(f, &(the_vm->activ_stack[i]),
		       i == 0 ? vector : MAIN_VECTOR) < 0) {
	    errlogf("READ_VM: Bad activ number %d\n", i);

	    for (j = 0; j < i; j++)
		free_activation(the_vm->activ_stack[i], 1 /* data_too */);
	    free_vm(the_vm, 0);
	    return -1;
	}

    *addr_for_the_vm = the_vm;    
    return 0;
}

char rcsid_eval_vm[] = "$Id: eval_vm.c,v 1.8 1993/08/11 02:58:03 pavel Exp $";

/* $Log: eval_vm.c,v $
 * Revision 1.8  1993/08/11  02:58:03  pavel
 * Changed some bogus %d's to %u's in calls to *scanf() and *printf(), guided
 * by warnings from GCC...
 *
 * Revision 1.7  1992/10/23  23:03:47  pavel
 * Added copyright notice.
 *
 * Revision 1.6  1992/10/21  03:02:35  pavel
 * Converted to use new automatic configuration system.
 *
 * Revision 1.5  1992/10/17  20:25:59  pavel
 * Changed return-type of read_vm() from `char' to `int', for systems that use
 * unsigned chars.
 *
 * Revision 1.4  1992/09/25  21:12:12  pjames
 * Passed correct pc (error_pc) to get_lineno.
 *
 * Revision 1.3  1992/09/02  18:43:11  pavel
 * Added mechanism for resuming a read()ing task with an error.
 *
 * Revision 1.2  1992/08/31  22:22:24  pjames
 * Changed some `char *'s to `const char *'
 *
 * Revision 1.1  1992/08/10  16:20:00  pjames
 * Initial RCS-controlled version
 */
