/******************************************************************************
  Copyright (c) 1992 Xerox Corporation.  All rights reserved.
  Portions of this code were written by Stephen White, aka ghond.
  Use and copying of this software and preparation of derivative works based
  upon this software are permitted.  Any distribution of this software or
  derivative works must comply with all applicable United States export
  control laws.  This software is made available AS IS, and Xerox Corporation
  makes no warranty about the software, its performance or its conformity to
  any specification.  Any person obtaining a copy of this software is requested
  to send their name and post office or electronic mail address to:
    Pavel Curtis
    Xerox PARC
    3333 Coyote Hill Rd.
    Palo Alto, CA 94304
    Pavel@Xerox.Com
 *****************************************************************************/

/* Multi-user networking protocol implementation for local clients on BSD UNIX
 */

#include <errno.h>		/* EMFILE */
#include "my-socket.h"		/* socket(), AF_UNIX, SOCK_STREAM,
				 * bind(), struct sockaddr, accept(),
				 * shutdown(), connect() */
#include "my-stdio.h"		/* remove() */
#include "my-string.h"		/* strcpy() */
#include <sys/un.h>		/* struct sockaddr_un */
#include "my-unistd.h"		/* close() */

#include "config.h"
#include "log.h"
#include "net_proto.h"

#define DEFAULT_CONNECT_FILE	"/tmp/.MOO-server"

static int server_socket;
static const char *connect_file;

const char *
proto_name(void)
{
    return "BSD single-host";
}

const char *
proto_usage_string(void)
{
    return "[server-connect-file]";
}

int
proto_initialize(struct proto *proto, int argc, char **argv)
{
    struct sockaddr_un	address;
    
    proto->pocket_size = 1;
    proto->believe_eof = 1;
    proto->eol_out_string = "\n";
    proto->eol_in_char = '\n';

    if (argc > 1)
	return 0;
    else if (argc == 1) {
	connect_file = argv[0];
    } else
	connect_file = DEFAULT_CONNECT_FILE;

    server_socket = socket(AF_UNIX, SOCK_STREAM, 0);
    if (server_socket < 0) {
	log_perror("Creating server socket");
	return 0;
    }

    address.sun_family = AF_UNIX;
    strcpy(address.sun_path, connect_file);
    if (bind(server_socket, (struct sockaddr *) &address,
	     sizeof(address.sun_family) + strlen(connect_file))) {
	log_perror("Binding server socket");
	close(server_socket);
	return 0;
    }
    proto->connection_fd = server_socket;
    return 1;
}

int
proto_listen(struct proto *proto)
{
    listen(server_socket, 5);
    return 1;
}

enum proto_accept_error
proto_accept_connection(int *read_fd, int *write_fd, const char **name)
{
    int				fd;
    static struct sockaddr_un	address;
    int				addr_length = sizeof(address);

    fd = accept(server_socket, (struct sockaddr *) &address, &addr_length);
    if (fd < 0) {
	if (errno == EMFILE)
	    return PA_FULL;
	else {
	    log_perror("Accepting new network connection");
	    return PA_OTHER;
	}
    }

    *read_fd = *write_fd = fd;
    *name = connect_file;
    return PA_OKAY;
}

void
proto_close_connection(int read_fd, int write_fd)
{
    /* read_fd and write_fd are the same, so we only need to deal with one. */
    shutdown(read_fd, 2);
    close(read_fd);
}

void
proto_shutdown(void)
{
    close(server_socket);
    remove(connect_file);
}

char rcsid_net_bsd_lcl[] = "$Id: net_bsd_lcl.c,v 1.6 1992/10/23 23:03:47 pavel Exp $";

/* $Log: net_bsd_lcl.c,v $
 * Revision 1.6  1992/10/23  23:03:47  pavel
 * Added copyright notice.
 *
 * Revision 1.5  1992/10/21  03:02:35  pavel
 * Converted to use new automatic configuration system.
 *
 * Revision 1.4  1992/10/06  01:36:23  pavel
 * Moved non-blocking code to net_multi.c, replacing it with code to set
 * proto->believe_eof appropriately.
 *
 * Revision 1.3  1992/09/26  02:25:02  pavel
 * Added support for printing the network protocol name on server start-up.
 *
 * Revision 1.2  1992/09/24  00:39:47  pavel
 * Removed useless #include's.
 *
 * Revision 1.1  1992/09/23  22:33:38  pavel
 * Initial RCS-controlled version.
 */
