/******************************************************************************
  Copyright (c) 1992 Xerox Corporation.  All rights reserved.
  Portions of this code were written by Stephen White, aka ghond.
  Use and copying of this software and preparation of derivative works based
  upon this software are permitted.  Any distribution of this software or
  derivative works must comply with all applicable United States export
  control laws.  This software is made available AS IS, and Xerox Corporation
  makes no warranty about the software, its performance or its conformity to
  any specification.  Any person obtaining a copy of this software is requested
  to send their name and post office or electronic mail address to:
    Pavel Curtis
    Xerox PARC
    3333 Coyote Hill Rd.
    Palo Alto, CA 94304
    Pavel@Xerox.Com
 *****************************************************************************/

/* Multi-user networking protocol implementation for TCP/IP on BSD UNIX */

#include "my-inet.h"		/* inet_addr() */
#include <errno.h>		/* EMFILE, EADDRNOTAVAIL, ECONNREFUSED,
				 * ENETUNREACH, ETIMEOUT */
#include "my-in.h"		/* struct sockaddr_in, INADDR_ANY, htons(),
				 * htonl(), ntohl(), struct in_addr */
#include "my-socket.h"		/* socket(), AF_INET, SOCK_STREAM,
				 * setsockopt(), SOL_SOCKET, SO_REUSEADDR,
				 * bind(), struct sockaddr, accept(),
				 * connect() */
#include "my-stdlib.h"		/* strtoul() */
#include "my-string.h"		/* memcpy() */
#include "my-unistd.h"		/* close() */

#include "config.h"
#include "log.h"
#include "name_lookup.h"
#include "net_proto.h"
#include "options.h"
#include "streams.h"
#include "timers.h"

#define DEFAULT_PORT 7777

static int server_socket;

const char *
proto_name(void)
{
    return "BSD/TCP";
}

const char *
proto_usage_string(void)
{
    return "[port]";
}

int
proto_initialize(struct proto *proto, int argc, char **argv)
{
    unsigned		port;
    char	       *p;
    struct sockaddr_in	address;
    int			option = 1;
    
    initialize_name_lookup();

    proto->pocket_size = 1;
    proto->believe_eof = 1;
    proto->eol_out_string = "\r\n";
    proto->eol_in_char = '\n';

    if (argc > 1)
	return 0;
    else if (argc == 1) {
	port = strtoul(argv[0], &p, 10);
	if (*p != '\0')
	    return 0;
    } else
	port = DEFAULT_PORT;

    server_socket = socket(AF_INET, SOCK_STREAM, 0);
    if (server_socket < 0) {
	log_perror("Creating server socket");
	return 0;
    }
    if (setsockopt(server_socket, SOL_SOCKET, SO_REUSEADDR,
		   (char *) &option, sizeof(option)) < 0) {
	log_perror("Setting server socket options");
	close(server_socket);
	return 0;
    }
    address.sin_family = AF_INET;
    address.sin_addr.s_addr = htonl(INADDR_ANY);
    address.sin_port = htons(port);
    if (bind(server_socket, (struct sockaddr *) &address, sizeof(address))) {
	log_perror("Binding server socket");
	close(server_socket);
	return 0;
    }
    proto->connection_fd = server_socket;
    return 1;
}

int
proto_listen(struct proto *proto)
{
    listen(server_socket, 5);
    return 1;
}

enum proto_accept_error
proto_accept_connection(int *read_fd, int *write_fd, const char **name)
{
    int			fd;
    struct sockaddr_in	address;
    int			addr_length = sizeof(address);
    static Stream      *s = 0;

    if (!s)
	s = new_stream(100);

    fd = accept(server_socket, (struct sockaddr *) &address, &addr_length);
    if (fd < 0) {
	if (errno == EMFILE)
	    return PA_FULL;
	else {
	    log_perror("Accepting new network connection");
	    return PA_OTHER;
	}
    }

    *read_fd = *write_fd = fd;
    stream_printf(s, "%s, port %d",
		  lookup_name_from_addr(&address, 5),
		  (int) ntohs(address.sin_port));
    *name = reset_stream(s);
    return PA_OKAY;
}

void
proto_close_connection(int read_fd, int write_fd)
{
    /* read_fd and write_fd are the same, so we only need to deal with one. */
    close(read_fd);
}

void
proto_shutdown(void)
{
    close(server_socket);
}

#ifdef OUTBOUND_NETWORK

#include "exceptions.h"
#include "structures.h"

static Exception	timeout;

static void
timeout_proc(Timer_ID id, Timer_Data data)
{
    RAISE(timeout, 0);
}

const char *
proto_outbound_name(Var arglist)
{
    /* The arglist is guaranteed proper for a call to proto_open_connection */
    static Stream      *s = 0;
    const char	       *host_name = arglist.v.list[1].v.str;
    unsigned	        port = arglist.v.list[2].v.num;

    if (!s)
	s = new_stream(100);

    stream_printf(s, "%s, port %d", host_name, port);

    return reset_stream(s);
}

enum error
proto_open_connection(Var arglist, int *read_fd, int *write_fd)
{
    /* These are `static' rather than `volatile' because I can't cope with
     * getting all those nasty little parameter-passing rules right.  This
     * function isn't recursive anyway, so it doesn't matter.
     */
    static const char		       *host_name;
    static unsigned			port;
    static struct sockaddr_in		addr;
    static Timer_ID			id;
    int					s, result;

    if (arglist.v.list[0].v.num != 2)
	return E_ARGS;
    else if (arglist.v.list[1].type != TYPE_STR ||
	     arglist.v.list[2].type != TYPE_NUM)
	return E_TYPE;

    host_name = arglist.v.list[1].v.str;
    port = arglist.v.list[2].v.num;

    addr.sin_family = AF_INET;
    addr.sin_port = htons(port);
    addr.sin_addr.s_addr = lookup_addr_from_name(host_name, 5);
    if (addr.sin_addr.s_addr == 0)
	return E_INVARG;

    s = socket(AF_INET, SOCK_STREAM, 0);
    if (s < 0) {
	if (errno != EMFILE)
	    log_perror("Making socket in proto_open_connection");
	return E_QUOTA;
    }

    TRY
	id = set_timer(5, timeout_proc, 0);
	result = connect(s, (struct sockaddr *) &addr, sizeof(addr));
        cancel_timer(id);
    EXCEPT (timeout)
	result = -1;
        errno = ETIMEDOUT;
	reenable_timers();
    ENDTRY

    if (result < 0) {
	close(s);
	if (errno == EADDRNOTAVAIL ||
	    errno == ECONNREFUSED  ||
	    errno == ENETUNREACH   ||
	    errno == ETIMEDOUT)
	    return E_INVARG;
	log_perror("Connecting in proto_open_connection");
	return E_QUOTA;
    }

    *read_fd = *write_fd = s;

    return E_NONE;
}
#endif /* OUTBOUND_NETWORK */

char rcsid_net_bsd_tcp[] = "$Id: net_bsd_tcp.c,v 1.13 1993/08/11 03:14:12 pavel Exp $";

/* $Log: net_bsd_tcp.c,v $
 * Revision 1.13  1993/08/11  03:14:12  pavel
 * -- Worked around a problem in some systems' declaration of gethostbyname().
 * -- Changed some bogus %d's to %u's in calls to *scanf() and *printf(),
 *    guided by warnings from GCC...
 *
 * Revision 1.12  1993/08/05  19:00:07  pavel
 * Made another local variable in proto_open_connection `static' to avoid
 * volatility warnings.
 *
 * Revision 1.11  1993/08/04  02:40:17  pavel
 * Sigh.  I never will learn not to check anything in until *after* it
 * compiles...
 *
 * Revision 1.10  1993/08/04  02:19:05  pavel
 * -- Added support for better logging of outbound connections.
 *
 * Revision 1.9  1993/08/04  00:25:54  pavel
 * Use strtoul() instead of strtol() for the port argument.
 *
 * Revision 1.8  1993/08/04  00:19:54  pavel
 * -- Fixed declaration of variables holding TCP ports to be unsigned.
 * -- Close server_socket if the setsockopt() fails.
 * -- Support new `network_listen()' protocol.
 * -- Set a timeout on outbound network connections.
 *
 * Revision 1.7  1992/10/23  23:03:47  pavel
 * Added copyright notice.
 *
 * Revision 1.6  1992/10/21  03:02:35  pavel
 * Converted to use new automatic configuration system.
 *
 * Revision 1.5  1992/10/17  20:43:13  pavel
 * Changed to use "my-inet.h" rather than <arpa/inet.h>.
 *
 * Revision 1.4  1992/10/06  01:36:45  pavel
 * Moved non-blocking code to net_multi.c, replacing it with code to set
 * proto->believe_eof appropriately.
 *
 * Revision 1.3  1992/10/04  04:58:03  pavel
 * Fixed #include's for declaration of inet_addr().  Also replaced INADDR_ANY
 * by htonl(INADDR_ANY), for portability.
 *
 * Revision 1.2  1992/09/26  02:22:53  pavel
 * Added support for printing the network protocol name on server start-up.
 *
 * Revision 1.1  1992/09/23  17:14:17  pavel
 * Initial RCS-controlled version.
 *
 */
