/******************************************************************************
  Copyright (c) 1992 Xerox Corporation.  All rights reserved.
  Portions of this code were written by Stephen White, aka ghond.
  Use and copying of this software and preparation of derivative works based
  upon this software are permitted.  Any distribution of this software or
  derivative works must comply with all applicable United States export
  control laws.  This software is made available AS IS, and Xerox Corporation
  makes no warranty about the software, its performance or its conformity to
  any specification.  Any person obtaining a copy of this software is requested
  to send their name and post office or electronic mail address to:
    Pavel Curtis
    Xerox PARC
    3333 Coyote Hill Rd.
    Palo Alto, CA 94304
    Pavel@Xerox.Com
 *****************************************************************************/

#include "my-ctype.h"
#include <errno.h>
#include "my-fcntl.h"
#include "my-ioctl.h"
#include "my-signal.h"
#include "my-stdio.h"
#include "my-stdlib.h"
#include "my-string.h"
#include "my-unistd.h"

#include "config.h"
#include "exceptions.h"
#include "log.h"
#include "net_proto.h"
#include "net_mplex.h"
#include "network.h"
#include "options.h"
#include "server.h"
#include "streams.h"
#include "structures.h"
#include "storage.h"
#include "timers.h"
#include "utils.h"

static struct proto 	proto;
static int		eol_length; /* == strlen(proto.eol_out_string) */

#ifdef EAGAIN
static int		eagain = EAGAIN;
#else
static int		eagain = -1;
#endif

#ifdef EWOULDBLOCK
static int		ewouldblock = EWOULDBLOCK;
#else
static int		ewouldblock = -1;
#endif

static int     *pocket_descriptors = 0;	/* fds we keep around in case we need
					 * one and no others are left... */
static const char *server_full_msg[] =
    {"*** Sorry, but the server cannot accept any more connections right now.",
     "*** Please try again later.",
     0};

typedef struct text_block {
    struct text_block  *next;
    unsigned		length;
    char	       *buffer;
    char	       *start;
} text_block;

typedef struct nhandle {
    struct nhandle     *next, **prev;
    server_handle	shandle;
    int			rfd, wfd;
    char	       *host_name;
    Stream	       *input;
    int			input_suspended;
    text_block	       *output_head;
    text_block	      **output_tail;
    unsigned		output_length;
    unsigned		output_lines_flushed;
    int			outbound;
} nhandle;

static nhandle *all_nhandles = 0;


static void
free_text_block(text_block *b)
{
    myfree(b->buffer, M_NETWORK);
    myfree(b, M_NETWORK);
}

int
set_non_blocking(int fd)
{
#ifdef FIONBIO
    /* Prefer this implementation, since the second one fails on some SysV
     * platforms, including HP/UX.
     */
    int	yes = 1;

    if (ioctl(fd, FIONBIO, &yes) < 0)
	return 0;
    else
	return 1;
#else
    int flags;

    if ((flags = fcntl(fd, F_GETFL, 0)) < 0
	|| fcntl(fd, F_SETFL, flags | NONBLOCK_FLAG) < 0)
	return 0;
    else
	return 1;
#endif
}

static int
push_output(nhandle *h)
{
    text_block *b;
    int		count;

    if (h->output_lines_flushed > 0) {
	char	buf[100];
	int	length;

	sprintf(buf,
		"%s>> Network buffer overflow: %u line%s of output to you %s been lost <<%s",
		proto.eol_out_string,
		h->output_lines_flushed,
		h->output_lines_flushed == 1 ? "" : "s",
		h->output_lines_flushed == 1 ? "has" : "have",
		proto.eol_out_string);
	length = strlen(buf);
	count = write(h->wfd, buf, length);
	if (count == length)
	    h->output_lines_flushed = 0;
	else
	    return count >= 0  ||  errno == eagain  ||  errno == ewouldblock;
    }
    while ((b = h->output_head) != 0) {
	count = write(h->wfd, b->start, b->length);
	if (count < 0)
	    return (errno == eagain  ||  errno == ewouldblock);
	h->output_length -= count;
	if (count == b->length) {
	    h->output_head = b->next;
	    free_text_block(b);
	} else {
	    b->start += count;
	    b->length -= count;
	}
    }
    if (h->output_head == 0)
	h->output_tail = &(h->output_head);
    return 1;
}

static int
pull_input(nhandle *h)
{
    Stream     *s = h->input;
    int		count;
    char	buffer[1024];
    char       *ptr, *end;

    if ((count = read(h->rfd, buffer, sizeof(buffer))) > 0) {
	for (ptr = buffer, end = buffer + count; ptr < end; ptr++) {
	    unsigned char	c = *ptr;

	    if (isgraph(c)  ||  c == ' '  ||  c == '\t')
		stream_add_char(s, c);
	    else if (c == proto.eol_in_char)
		server_receive_line(h->shandle, reset_stream(s));
	}
	return 1;
    } else
	return (count == 0  &&  !proto.believe_eof)
	    || (count < 0  &&  (errno == eagain  ||  errno == ewouldblock));
}

static void
make_new_connection(int rfd, int wfd, const char *host_name, int outbound)
{
    nhandle	       *h;
    network_handle	nh;

    if (!set_non_blocking(rfd)
	|| (rfd != wfd  &&  !set_non_blocking(wfd)))
	log_perror("Setting connection non-blocking");
    h = mymalloc(sizeof(nhandle), M_NETWORK);
    if (all_nhandles)
	all_nhandles->prev = &(h->next);
    h->next = all_nhandles;
    h->prev = &all_nhandles;
    all_nhandles = h;
    h->rfd = rfd;
    h->wfd = wfd;
    h->input = new_stream(100);
    h->input_suspended = 0;
    h->output_head = 0;
    h->output_tail = &(h->output_head);
    h->output_length = 0;
    h->output_lines_flushed = 0;
    h->host_name = (char *) mymalloc(strlen(host_name) + 1, M_NETWORK);
    strcpy(h->host_name, host_name);
    h->outbound = outbound;

    nh.ptr = h;
    h->shandle = server_new_connection(nh, outbound);
}

static void
get_pocket_descriptors()
{
    int	i;

    if (!pocket_descriptors)
	pocket_descriptors =
	    (int *) mymalloc(proto.pocket_size * sizeof(int), M_NETWORK);

    for (i = 0; i < proto.pocket_size; i++) {
	pocket_descriptors[i] = dup(0);
	if (!pocket_descriptors[i]) {
	    log_perror("Can't get a pocket descriptor");
	    panic("Need pocket descriptors to continue");
	}
    }
}

static void
accept_new_connection(void)
{
    int		rfd, wfd, i;
    const char *host_name;

    switch (proto_accept_connection(&rfd, &wfd, &host_name)) {
      case PA_OKAY:
	make_new_connection(rfd, wfd, host_name, 0);
	break;

      case PA_FULL:
	for (i = 0; i < proto.pocket_size; i++)
	    close(pocket_descriptors[i]);
	if (proto_accept_connection(&rfd, &wfd, &host_name) != PA_OKAY)
	    errlogf("Can't accept connection even by emptying pockets!\n");
	else {
	    for (i = 0; server_full_msg[i] != 0; i++) {
		write(wfd, server_full_msg[i], strlen(server_full_msg[i]));
		write(wfd, proto.eol_out_string, eol_length);
	    }
	    logf("SERVER FULL: refusing connection from %s\n", host_name);
	    proto_close_connection(rfd, wfd);
	}
	get_pocket_descriptors();
	break;

      case PA_OTHER:
	/* Do nothing.  The protocol implementation has already logged it. */
	break;
    }
}

static int
enqueue_output(network_handle nh, const char *line, int add_eol, int flush_ok)
{
    nhandle    *h = nh.ptr;
    unsigned	line_length = strlen(line);
    unsigned	length = line_length + (add_eol ? eol_length : 0);
    char       *buffer;
    text_block *block;

    if (h->output_length + length > MAX_QUEUED_OUTPUT) {
	int		to_flush;
	text_block     *b;
	
	(void) push_output(h);
	to_flush = h->output_length + length - MAX_QUEUED_OUTPUT;
	if (to_flush > 0 && !flush_ok)
	    return 0;
	while (to_flush > 0  &&  (b = h->output_head)) {
	    h->output_length -= b->length;
	    to_flush -= b->length;
	    h->output_lines_flushed++;
	    h->output_head = b->next;
	    free_text_block(b);
	}
	if (h->output_head == 0)
	    h->output_tail = &(h->output_head);
    }

    buffer = (char *) mymalloc(length * sizeof(char), M_NETWORK);
    block = (text_block *) mymalloc(sizeof(text_block), M_NETWORK);
    memcpy(buffer, line, line_length);
    if (add_eol)
	memcpy(buffer + line_length, proto.eol_out_string, eol_length);
    block->buffer = block->start = buffer;
    block->length = length;
    block->next = 0;
    *(h->output_tail) = block;
    h->output_tail = &(block->next);
    h->output_length += length;

    return 1;
}

static void
close_nhandle(nhandle *h)
{
    text_block *b, *bb;

    (void) push_output(h);
    *(h->prev) = h->next;
    if (h->next)
	h->next->prev = h->prev;
    b = h->output_head;
    while (b) {
	bb = b->next;
	free_text_block(b);
	b = bb;
    }
    free_stream(h->input);
    proto_close_connection(h->rfd, h->wfd);
    myfree(h->host_name, M_NETWORK);
    myfree(h, M_NETWORK);
}


/*************************
 * External entry points *
 *************************/

const char *
network_protocol_name(void)
{
    return proto_name();
}

const char *
network_usage_string(void)
{
    return proto_usage_string();
}

int
network_initialize(int argc, char **argv)
{
    if (!proto_initialize(&proto, argc, argv))
	return 0;

    eol_length = strlen(proto.eol_out_string);
    get_pocket_descriptors();

    /* we don't care about SIGPIPE, we notice it in mplex_wait() and write() */
    signal(SIGPIPE, SIG_IGN);

    return 1;
}

int
network_listen(void)
{
    return proto_listen(&proto);
}

int
network_send_line(network_handle nh, const char *line, int flush_ok)
{
    return enqueue_output(nh, line, 1, flush_ok);
}

void
network_suspend_input(network_handle nh)
{
    nhandle    *h = nh.ptr;

    h->input_suspended = 1;
}

void
network_resume_input(network_handle nh)
{
    nhandle    *h = nh.ptr;

    h->input_suspended = 0;
}

int
network_process_io(unsigned timeout)
{
    nhandle    *h, *hnext;

    mplex_clear();
    mplex_add_reader(proto.connection_fd);
    for (h = all_nhandles; h; h = h->next) {
	if (!h->input_suspended)
	    mplex_add_reader(h->rfd);
	if (h->output_head)
	    mplex_add_writer(h->wfd);
    }

    if (mplex_wait(timeout))
	return 0;
    else {
	if (mplex_is_readable(proto.connection_fd))
	    accept_new_connection();
	for (h = all_nhandles; h; h = hnext) {
	    hnext = h->next;
	    if ((mplex_is_readable(h->rfd)  &&  !pull_input(h))
		|| (mplex_is_writable(h->wfd)  &&  !push_output(h))) {
		server_close(h->shandle);
		close_nhandle(h);
	    }
	}
	return 1;
    }
}

const char *
network_connection_name(network_handle nh)
{
    nhandle	       *h = (nhandle *) nh.ptr;
    static Stream      *s = 0;

    if (s == 0)
	s = new_stream(100);

    if (h->rfd == h->wfd)
	stream_printf(s, "%d", h->rfd);
    else
	stream_printf(s, "(%d,%d)", h->rfd, h->wfd);
    stream_printf(s, " %s %s", h->outbound ? "to" : "from", h->host_name);
    return reset_stream(s);
}

int
network_set_connection_option(network_handle nh, const char *option, Var value)
{
#if NETWORK_PROTOCOL == NP_TCP
	/* These values taken from RFC 854 and RFC 857. */
#define TN_IAC	255		/* Interpret As Command */
#define TN_WILL	251
#define TN_WONT	252
#define TN_ECHO	1

    static char	telnet_cmd[4] = {TN_IAC, 0, TN_ECHO, 0};

    if (!mystrcasecmp(option, "client-echo")) {
	if (is_true(value))
	    telnet_cmd[1] = TN_WONT;
	else
	    telnet_cmd[1] = TN_WILL;
	enqueue_output(nh, telnet_cmd, 0, 1);
	return 1;
    } else
#endif
	return 0;
}

#ifdef OUTBOUND_NETWORK
#include "structures.h"

enum error
network_open_connection(Var arglist)
{
    int		rfd, wfd;
    enum error	e;

    e = proto_open_connection(arglist, &rfd, &wfd);
    if (e == E_NONE)
	make_new_connection(rfd, wfd, proto_outbound_name(arglist), 1);

    return e;
}
#endif

void
network_close(network_handle h)
{
    close_nhandle((nhandle *) h.ptr);
}

void
network_shutdown(void)
{
    while (all_nhandles)
	close_nhandle(all_nhandles);
    proto_shutdown();
}

char rcsid_net_multi[] = "$Id: net_multi.c,v 1.16 1993/08/12 21:10:13 pavel Exp $";

/* $Log: net_multi.c,v $
 * Revision 1.16  1993/08/12  21:10:13  pavel
 * Fix long-standing denial-of-service attack vulnerability due to a connection
 * sending an essentially infinite stream of input to the server.
 *
 * Revision 1.15  1993/08/11  03:11:17  pavel
 * -- Fixed a syntax error in the new support for outbound connections.
 * -- Changed some bogus %d's to %u's in calls to *scanf() and *printf(),
 *    guided by warnings from GCC...
 *
 * Revision 1.14  1993/08/04  02:21:48  pavel
 * -- Added support for distinguishing between inbound and outbound
 *    connections.
 * -- Added support for better logging of outbound connections.
 * -- Added check to connection-timeout code to exempt outbound connections.
 *
 * Revision 1.13  1993/08/04  01:33:04  pavel
 * -- Added support for the BSD-style ioctl(fd, FIONBIO, ...) non-blocking I/O
 *    protocol.  The SysV/POSIX-style fcntl(fd, F_GETFL, ...) doesn't work on
 *    BSD sockets on some hybrid systems, including HP/UX.
 * -- Vastly improved the clarity of the `output flushed' message users see
 *    when their output-side network buffer overflows.
 * -- Now accepts tab as a normal input character mapping to itself, necessary
 *    for communicating with Gopher from the MOO.
 * -- Fixed a =/== bug in pull_input().
 * -- Added a log message printed when the server has to refuse a connection
 *    because there aren't any file descriptors left.
 * -- Added support for the new network_listen() protocol.
 *
 * Revision 1.12  1992/10/23  23:03:47  pavel
 * Added copyright notice.
 *
 * Revision 1.11  1992/10/21  03:02:35  pavel
 * Converted to use new automatic configuration system.
 *
 * Revision 1.10  1992/10/17  20:44:14  pavel
 * Changed to use NONBLOCK_FLAG instead of O_NDELAY to allow of using
 * POSIX-style non-blocking on systems where it is available.
 *
 * Revision 1.9  1992/10/06  18:16:36  pavel
 * Moved non-blocking code to here from individual protocol implementations.
 *
 * Revision 1.8  1992/09/30  06:18:08  pavel
 * Fixed small bug in the handling of the case where even emptying our pockets
 * of extra file descriptors doesn't make it possible to accept a connection.
 *
 * Revision 1.7  1992/09/26  02:22:07  pavel
 * Added support for printing the network protocol name on server start-up.
 *
 * Revision 1.6  1992/09/23  17:17:36  pavel
 * Stripped out all BSD-specific code, instead relying on the net_proto.h
 * and net_mplex.h interfaces.
 *
 * Revision 1.5  1992/09/11  21:21:54  pavel
 * Tracked change to network.h.
 *
 * Revision 1.4  1992/09/04  22:41:47  pavel
 * Fixed some picky ANSI C problems with (const char *)'s.
 *
 * Revision 1.3  1992/08/10  17:22:45  pjames
 * Updated #includes.
 *
 * Revision 1.2  1992/07/20  23:56:16  pavel
 * Added rcsid_<filename-root> declaration to hold the RCS ident. string.
 *
 * Revision 1.1  1992/07/20  23:23:12  pavel
 * Initial RCS-controlled version.
 */
