/******************************************************************************
  Copyright (c) 1992 Xerox Corporation.  All rights reserved.
  Portions of this code were written by Stephen White, aka ghond.
  Use and copying of this software and preparation of derivative works based
  upon this software are permitted.  Any distribution of this software or
  derivative works must comply with all applicable United States export
  control laws.  This software is made available AS IS, and Xerox Corporation
  makes no warranty about the software, its performance or its conformity to
  any specification.  Any person obtaining a copy of this software is requested
  to send their name and post office or electronic mail address to:
    Pavel Curtis
    Xerox PARC
    3333 Coyote Hill Rd.
    Palo Alto, CA 94304
    Pavel@Xerox.Com
 *****************************************************************************/

/* This describes the complete set of procedures that a multi-user network
 * protocol implementation must provide.
 */

#ifndef Net_Proto_H
#define Net_Proto_H 1

#include "config.h"
#include "options.h"

struct proto {
    unsigned	pocket_size;	/* Maximum number of file descriptors it might
				 * take to accept a new connection in this
				 * protocol.  The generic multi-user network
				 * code will keep this many descriptors `in its
				 * pocket', ready to be freed up in order to
				 * tell potential users that there's no more
				 * room in the server. */
    int		believe_eof;	/* If true, then read() will return 0 on a
				 * connection using this protocol iff the
				 * connection really is closed.  If false, then
				 * read() -> 0 will be interpreted as the
				 * connection still being open, but no data
				 * being available. */
    int		connection_fd;	/* An open file descriptor that will be ready
				 * ready for input operations iff a new
				 * connection is available. */
    const char *eol_out_string;	/* The characters to add to the end of each
				 * line of output on connections. */
    unsigned char eol_in_char;	/* A character that will appear at the end of
				 * every line of input on connections. */
};

extern const char      *proto_name(void);
				/* Returns a string naming the protocol. */

extern const char      *proto_usage_string(void);
				/* Returns a string giving the syntax of any
				 * extra, protocol-specific command-line
				 * arguments, such as a port number. */

extern int		proto_initialize(struct proto *proto,
					 int argc, char **argv);
				/* argc and argv refer to just the protocol-
				 * specific command-line arguments, if any,
				 * which always come after any protocol-
				 * independent args.  Returns true iff those
				 * arguments were valid and network
				 * initialization succeeded.  On success, all
				 * of the fields of proto should be filled in
				 * with values appropriate for the protocol.
				 *
				 * NOTE: It is okay for connection_fd not to
				 * be filled in on return from this call, and
				 * more than okay for the server still to be
				 * refusing connections.  The server's call to
				 * proto_listen() marks the time by which
				 * connection_fd must exist and the server must
				 * start accepting connections.
  				 */

extern int		proto_listen(struct proto *proto);
				/* The protocol implementation should do any
				 * work necessary to finish initializing
				 * `proto' and to prepare for accepting
				 * connections, returning true iff it is
				 * successful.  When this returns true,
				 * `proto->connection_fd' must have been filled
				 * in with a value satisfying the specification
				 * given above.
				 */


enum proto_accept_error { PA_OKAY, PA_FULL, PA_OTHER };

extern enum proto_accept_error
	    		proto_accept_connection(int *read_fd, int *write_fd,
						const char **name);
				/* Accept a new connection, returning PA_OKAY
				 * if successful, PA_FULL is unsuccessful only
				 * because there aren't enough file descriptors
				 * available, and PA_OTHER for other failures
				 * (in which case a message should have been
				 * output to the server log).  On successful
				 * return, *read_fd and *write_fd should be
				 * file descriptors on which input and output
				 * for the new connection can be done, and
				 * *name should be a human-readable string
				 * identifying this connection.
				 */

extern void		proto_close_connection(int read_fd, int write_fd);
				/* Close the given file descriptors, which were
				 * returned by proto_accept_connection(),
				 * performing whatever extra clean-ups are
				 * required by the protocol.
				 */

#ifdef OUTBOUND_NETWORK
#include "structures.h"

extern enum error	proto_open_connection(Var arglist,
					      int *read_fd, int *write_fd);
				/* The given MOO arguments should be used as a
				 * specification of a remote network connection
				 * to be opened.  If the arguments are OK for
				 * this protocol and the connection is success-
				 * fully made, then *read_fd and *write_fd
				 * should be set as proto_accept_connection()
				 * does and E_NONE returned.  Otherwise, an
				 * appropriate error should be returned.
				 */

extern const char      *proto_outbound_name(Var arglist);
				/* The given MOO arguments were used in a
				 * successful call to proto_open_connection().
				 * Return a human-readable string summarizing
				 * their contents.
				 */

#endif /* OUTBOUND_NETWORK */

extern void		proto_shutdown(void);
				/* As of this call, the generic network code
				 * will never make another call on the
				 * protocol; any necessary clean-ups should
				 * be done, such as closing the protocol's
				 * connection_fd.
				 */

#endif /* !Net_Proto_H */

/* $Log: net_proto.h,v $
 * Revision 1.6  1993/08/04  02:07:23  pavel
 * -- Added support for the new network_listen() protocol.
 * -- Added support for better logging of outbound connections.
 *
 * Revision 1.5  1992/10/23  23:03:47  pavel
 * Added copyright notice.
 *
 * Revision 1.4  1992/10/21  03:02:35  pavel
 * Converted to use new automatic configuration system.
 *
 * Revision 1.3  1992/10/06  18:18:04  pavel
 * Replaced struct proto members empty_read_errno and full_write_errno by
 * believe_eof, as part of the move of the non-blocking code from protocol
 * implementations to net_multi.c.
 *
 * Revision 1.2  1992/09/26  02:24:07  pavel
 * Added support for printing the network protocol name on server start-up.
 *
 * Revision 1.1  1992/09/23  17:14:17  pavel
 * Initial RCS-controlled version.
 */
