/******************************************************************************
  Copyright (c) 1992 Xerox Corporation.  All rights reserved.
  Portions of this code were written by Stephen White, aka ghond.
  Use and copying of this software and preparation of derivative works based
  upon this software are permitted.  Any distribution of this software or
  derivative works must comply with all applicable United States export
  control laws.  This software is made available AS IS, and Xerox Corporation
  makes no warranty about the software, its performance or its conformity to
  any specification.  Any person obtaining a copy of this software is requested
  to send their name and post office or electronic mail address to:
    Pavel Curtis
    Xerox PARC
    3333 Coyote Hill Rd.
    Palo Alto, CA 94304
    Pavel@Xerox.Com
 *****************************************************************************/

/* Multi-user networking protocol implementation for TCP/IP on System V */

/* It appears that there may never be a use for this particular protocol
 * implementation because every vendor that provides a TLI interface to TCP
 * also provides a complete BSD-style sockets implementation, so one could
 * simply use net_bsd_tcp.c instead.  I went ahead and provided this code
 * anyway, for two reasons: (a) it gave the server code a kind of pleasing
 * symmetry and completeness (and me a sense of closure), and (b) it can,
 * perhaps, serve as a model for other TLI-based protocol implementations in
 * the future.  For example, there are vendors supplying TLI interfaces to the
 * XNS protocol family without providing a sockets-based interface.  In any
 * case, it was interesting to learn about TLI and may prove useful in the
 * future; I have been given to understand that TLI will probably form the
 * basis for the eventual POSIX standard networking interface.
 */

#include "my-inet.h"		/* inet_addr() */
#include <errno.h>		/* EMFILE */
#include "my-fcntl.h"		/* O_RDWR */
#include "my-in.h"		/* struct sockaddr_in, INADDR_ANY, htonl(),
				 * htons(), ntohl(), struct in_addr */
#include "my-ioctl.h"		/* ioctl() */
#include "my-socket.h"		/* AF_INET */
#include "my-stdlib.h"		/* strtoul() */
#include "my-string.h"		/* memcpy() */
#include "my-stropts.h"		/* I_POP, I_PUSH */
#include <sys/conf.h>		/* FMNAMESZ */
#include "my-tiuser.h"		/* t_open(), t_bind(), t_alloc(), t_accept() */
#include "my-unistd.h"		/* close() */

#include "config.h"
#include "log.h"
#include "name_lookup.h"
#include "net_proto.h"
#include "options.h"
#include "streams.h"
#include "timers.h"

#define DEFAULT_PORT 7777

static int 		server_fd;
static struct t_call   *call;

static void
log_ti_error(const char *msg)
{
    if (t_errno == TSYSERR)
	log_perror(msg);
    else
	errlogf("%s: %s\n", msg, t_errlist[t_errno]);
}

const char *
proto_name(void)
{
    return "SysV/TCP";
}

const char *
proto_usage_string(void)
{
    return "[port]";
}

int
proto_initialize(struct proto *proto, int argc, char **argv)
{
    unsigned		port;
    char	       *p;
    struct sockaddr_in	req_addr, rec_addr;
    struct t_bind	requested, received;

    initialize_name_lookup();
    
    proto->pocket_size = 1;
    proto->believe_eof = 1;
    proto->eol_out_string = "\r\n";
    proto->eol_in_char = '\n';

    if (argc > 1)
	return 0;
    else if (argc == 1) {
	port = strtoul(argv[0], &p, 10);
	if (*p != '\0')
	    return 0;
    } else
	port = DEFAULT_PORT;

    server_fd = t_open((void *) "/dev/tcp", O_RDWR, 0);
    if (server_fd < 0) {
	log_ti_error("Creating server endpoint");
	return 0;
    }

    req_addr.sin_family = AF_INET;
    req_addr.sin_addr.s_addr = htonl(INADDR_ANY);
    req_addr.sin_port = htons(port);

    requested.addr.maxlen = sizeof(req_addr);
    requested.addr.len = sizeof(req_addr);
    requested.addr.buf = (void *) &req_addr;
    requested.qlen = 5;

    received.addr.maxlen = sizeof(rec_addr);
    received.addr.len = sizeof(rec_addr);
    received.addr.buf = (void *) &rec_addr;

    if (t_bind(server_fd, &requested, &received) < 0) {
	log_ti_error("Binding to server address");
	t_close(server_fd);
	return 0;
    } else if (rec_addr.sin_port != port) {
	errlogf("Can't bind to requested port!\n");
	t_close(server_fd);
	return 0;
    }
    
    call = (struct t_call *) t_alloc(server_fd, T_CALL, T_ADDR);
    if (!call) {
	log_ti_error("Allocating T_CALL structure");
	t_close(server_fd);
	return 0;
    }

    proto->connection_fd = server_fd;
    return 1;
}

int
proto_listen(struct proto *proto)
{
    /* Nothing yet. */
    return 1;
}

static int
set_rw_able(int fd)
{
    char	buffer[FMNAMESZ + 1];

    if (ioctl(fd, I_LOOK, buffer) < 0
	|| strcmp(buffer, "timod") != 0
	|| ioctl(fd, I_POP, 0) < 0
	|| ioctl(fd, I_PUSH, (void *) "tirdwr") < 0) {
	log_perror("Setting up read/write behavior for connection");
	return 0;
    } else
	return 1;
}

enum proto_accept_error
proto_accept_connection(int *read_fd, int *write_fd, const char **name)
{
    int			fd;
    struct sockaddr_in *addr = (struct sockaddr_in *) call->addr.buf;
    static Stream      *s = 0;

    if (!s)
	s = new_stream(100);

    fd = t_open((void *) "/dev/tcp", O_RDWR, 0);
    if (fd < 0) {
	if (t_errno == TSYSERR  &&  errno == EMFILE)
	    return PA_FULL;
	else {
	    log_ti_error("Opening endpoint for new connection");
	    return PA_OTHER;
	}
    }

    if (t_bind(fd, 0, 0) < 0) {
	log_ti_error("Opening endpoint for new connection");
	t_close(fd);
	return PA_OTHER;
    }

    if (t_listen(server_fd, call) < 0) {
	log_ti_error("Accepting new network connection");
	t_close(fd);
	return PA_OTHER;
    }

    if (t_accept(server_fd, fd, call) < 0) {
	log_ti_error("Accepting new network connection");
	t_close(fd);
	return PA_OTHER;
    }

    if (!set_rw_able(fd)) {
	t_close(fd);
	return PA_OTHER;
    }

    *read_fd = *write_fd = fd;
    stream_printf(s, "%s, port %d",
		  lookup_name_from_addr(addr, 5),
		  (int) ntohs(addr->sin_port));
    *name = reset_stream(s);
    return PA_OKAY;
}

void
proto_close_connection(int read_fd, int write_fd)
{
    /* read_fd and write_fd are the same, so we only need to deal with one. */
    t_close(read_fd);
}

void
proto_shutdown(void)
{
    t_close(server_fd);
}

#ifdef OUTBOUND_NETWORK

#include "exceptions.h"
#include "structures.h"

static Exception	timeout;

static void
timeout_proc(Timer_ID id, Timer_Data data)
{
    RAISE(timeout, 0);
}

const char *
proto_outbound_name(Var arglist)
{
    /* The arglist is guaranteed proper for a call to proto_open_connection */
    static Stream      *s = 0;
    const char	       *host_name = arglist.v.list[1].v.str;
    unsigned	        port = arglist.v.list[2].v.num;

    if (!s)
	s = new_stream(100);

    stream_printf(s, "%s, port %d", host_name, port);

    return reset_stream(s);
}

enum error
proto_open_connection(Var arglist, int *read_fd, int *write_fd)
{
    /* These are `static' rather than `volatile' because I can't cope with
     * getting all those nasty little parameter-passing rules right.  This
     * function isn't recursive anyway, so it doesn't matter.
     */
    static const char	       *host_name;
    static unsigned		port;
    static Timer_ID		id;
    int				fd, result;
    static struct sockaddr_in	addr;

    if (arglist.v.list[0].v.num != 2)
	return E_ARGS;
    else if (arglist.v.list[1].type != TYPE_STR ||
	     arglist.v.list[2].type != TYPE_NUM)
	return E_TYPE;

    host_name = arglist.v.list[1].v.str;
    port = arglist.v.list[2].v.num;

    addr.sin_family = AF_INET;
    addr.sin_port = htons(port);
    addr.sin_addr.s_addr = lookup_addr_from_name(host_name, 5);
    if (addr.sin_addr.s_addr == 0)
	return E_INVARG;

    /* Cast to (void *) here to workaround const-less decls on some systems. */
    fd = t_open((void *) "/dev/tcp", O_RDWR, 0);
    if (fd < 0) {
	if (t_errno != TSYSERR  ||  errno != EMFILE)
	    log_ti_error("Making endpoint in proto_open_connection");
	return E_QUOTA;
    }

    if (t_bind(fd, 0, 0) < 0) {
	log_ti_error("Binding outbound endpoint");
	t_close(fd);
	return E_QUOTA;
    }

    call->addr.maxlen = sizeof(addr);
    call->addr.len = sizeof(addr);
    call->addr.buf = (void *) &addr;

    TRY
	id = set_timer(5, timeout_proc, 0);
	result = t_connect(fd, call, 0);
        cancel_timer(id);
    EXCEPT (timeout)
	result = -1;
        errno = ETIMEDOUT;
	t_errno = TSYSERR;
	reenable_timers();
    ENDTRY

    if (result < 0) {
	t_close(fd);
	log_ti_error("Connecting in proto_open_connection");
	return E_QUOTA;
    }

    if (!set_rw_able(fd)) {
	t_close(fd);
	return E_QUOTA;
    }

    *read_fd = *write_fd = fd;

    return E_NONE;
}
#endif /* OUTBOUND_NETWORK */

char rcsid_net_sysv_tcp[] = "$Id: net_sysv_tcp.c,v 1.11 1993/10/11 19:02:14 pavel Exp $";

/* $Log: net_sysv_tcp.c,v $
 * Revision 1.11  1993/10/11  19:02:14  pavel
 * Worked around problem with a mis-declared t_open() on Solaris, where they
 * neglected to specify `const' on the first argument.
 *
 * Revision 1.10  1993/08/11  03:15:44  pavel
 * -- Worked around a problem in some systems' declaration of gethostbyname().
 * -- Changed some bogus %d's to %u's in calls to *scanf() and *printf(),
 *    guided by warnings from GCC...
 *
 * Revision 1.9  1993/08/05  19:01:03  pavel
 * Made another local variable in proto_open_connection `static' to avoid
 * volatility warnings.
 *
 * Revision 1.8  1993/08/04  02:41:03  pavel
 * Sigh.  I never will learn not to check anything in until *after* it
 * compiles...
 *
 * Revision 1.7  1993/08/04  02:19:58  pavel
 * -- Added support for better logging of outbound connections.
 *
 * Revision 1.6  1993/08/04  00:25:10  pavel
 * Use strtoul() instead of strtol() for the port argument.
 *
 * Revision 1.5  1993/08/04  00:22:12  pavel
 * -- Fixed declaration of variables holding TCP ports to be unsigned.
 * -- Support new `network_listen()' protocol.
 * -- Set a timeout on outbound network connections.
 *
 * Revision 1.4  1992/10/23  23:03:47  pavel
 * Added copyright notice.
 *
 * Revision 1.3  1992/10/21  03:02:35  pavel
 * Converted to use new automatic configuration system.
 *
 * Revision 1.2  1992/10/17  21:02:51  pavel
 * Changed to use "my-inet.h" rather than <arpa/inet.h>.
 * Fixed typo in OUTBOUND_NETWORK support.
 *
 * Revision 1.1  1992/10/06  01:38:46  pavel
 * Initial RCS-controlled version.
 */
