/******************************************************************************
  Copyright (c) 1992 Xerox Corporation.  All rights reserved.
  Portions of this code were written by Stephen White, aka ghond.
  Use and copying of this software and preparation of derivative works based
  upon this software are permitted.  Any distribution of this software or
  derivative works must comply with all applicable United States export
  control laws.  This software is made available AS IS, and Xerox Corporation
  makes no warranty about the software, its performance or its conformity to
  any specification.  Any person obtaining a copy of this software is requested
  to send their name and post office or electronic mail address to:
    Pavel Curtis
    Xerox PARC
    3333 Coyote Hill Rd.
    Palo Alto, CA 94304
    Pavel@Xerox.Com
 *****************************************************************************/

#include "my-ctype.h"
#include "my-stdio.h"
#include "my-stdlib.h"
#include "my-string.h"
#include "my-time.h"

#include "config.h"
#include "db.h"
#include "list.h"
#include "match.h"
#include "parse_cmd.h"
#include "storage.h"
#include "structures.h"
#include "utils.h"

static char **
parse_into_words(char *input, int *nwords)
{
    static char	      **words = 0;
    static int		max_words = 0;
    int			in_quotes = 0;
    char	       *ptr = input;

    if (!words) {
	max_words = 50;
	words = mymalloc(max_words * sizeof(char *), M_STRING);
    }

    while (*input == ' ')
	input++;

    for (*nwords = 0; *input != '\0'; (*nwords)++) {
	if (*nwords == max_words) {
	    int		new_max = max_words * 2;
	    char      **new = mymalloc(new_max * sizeof(char *), M_STRING);
	    int		i;

	    for (i = 0; i < max_words; i++)
		new[i] = words[i];

	    myfree(words, M_STRING);
	    words = new;
	    max_words = new_max;
	}

	words[*nwords] = ptr;
	while (*input != '\0'  &&  (in_quotes  ||  *input != ' ')) {
	    char	c = *(input++);

	    if (c == '"')
		in_quotes = !in_quotes;
	    else if (c == '\\') {
		if (*input != '\0')
		    *(ptr++) = *(input++);
	    } else
		*(ptr++) = c;
	}
	while (*input == ' ')
	    input++;
	*(ptr++) = '\0';
    }

    return words;
}

static char *
build_string(int argc, char *argv[])
{
    int         i, len = 0;
    char	*str;

    if (!argc)
        return str_dup("");

    len = strlen(argv[0]);
    for (i = 1; i < argc; i++)
	len += 1 + strlen(argv[i]);

    str = (char *) mymalloc(len + 1, M_STRING);

    strcpy(str, argv[0]);
    for (i = 1; i < argc; i++) {
	strcat(str, " ");
	strcat(str, argv[i]);
    }

    return str;
}

/* Prepositions */

#define MAXPPHRASE 5	/* max. number of words in a prepositional phrase */
static const char  *prep_list[] = {
    "with/using",
    "at/to",
    "in front of",
    "in/inside/into",
    "on top of/on/onto/upon",
    "out of/from inside/from",
    "over",
    "through",
    "under/underneath/beneath",
    "behind",
    "beside",
    "for/about",
    "is",
    "as",
    "off/off of",
};

#define NPREPS Arraysize(prep_list, const char *)

struct  pt_entry {
    int		nwords;
    char	*words[MAXPPHRASE];
    struct pt_entry *next;
};

struct pt_entry	*prep_table[NPREPS];

void
build_preptable(void)
{
    int		i, j;
    int		nwords;
    char      **words;
    char	cprep[100];
    const char	*p;
    char	*t;
    struct pt_entry	*current_alias, **prev;

    for (i = 0; i < NPREPS; i++) {
	p = prep_list[i];
	prev = &(prep_table[i]);
	while (*p) {
	    t = cprep;
	    if (*p == '/')
		p++;
	    while (*p && *p != '/')
		*t++ = *p++;
	    *t = '\0';

	    words = parse_into_words(cprep, &nwords);

	    current_alias = (struct pt_entry *)
			    mymalloc(sizeof(struct pt_entry), M_PREP);
	    current_alias->nwords = nwords;
	    current_alias->next = 0;
	    for (j = 0; j < nwords; j++)
		current_alias->words[j] = str_dup(words[j]);
	    if (prev) 
		*prev = current_alias;
	    prev = &current_alias->next;
	}
    }
}

static Prepid
prep_find(int argc, char *argv[], int *pstart, int *pend)
{
    struct pt_entry	*alias;
    int			i, j, k;
    int			exact_match = (pstart == 0  ||  pend == 0);

    for (i = 0; i < argc; i++) {
	for (j = 0; j < NPREPS; j++) {
	    for (alias = prep_table[j]; alias; alias = alias->next) {
		if (i + alias->nwords <= argc) {
		    for (k = 0; k < alias->nwords; k++) {
			if (mystrcasecmp(argv[i + k], alias->words[k]))
			    break;
		    }
		    if (k == alias->nwords
			     &&  (! exact_match || i + k == argc)) {
			if (! exact_match) {
			    *pstart = i;
			    *pend = i + alias->nwords - 1;
			}
			return j;
		    }
		}
	    }
	}
	if (exact_match)
	    break;
    }
    return FAILED_MATCH;
}

Prepid
match_prep(const char *prepname)
{
    Prepid	prep;
    int		argc;
    char       *ptr;
    char      **argv;
    char       *s, first;

    if (!mystrcasecmp(prepname, "none")) return NOTHING;
    if (!mystrcasecmp(prepname, "any")) return AMBIGUOUS;

    s = str_dup(prepname);
    first = s[0];
    if (first == '#')
	s++;
    prep = strtol(s, &ptr, 10);
    if (*ptr == '\0') {
	free_str(s);
	if (!isdigit(first) || prep >= NPREPS)
	    return FAILED_MATCH;
	else
	    return prep;
    }

    if ((ptr = strchr(s, '/')) != '\0')
	*ptr = '\0';

    argv = parse_into_words(s, &argc);
    prep = prep_find(argc, argv, 0, 0);
    free_str(s);
    return prep;
}

const char *
prep_name(Prepid prep)
{
    return prep_list[prep];
}

/* Verb lookup */

Program *
find_verb_on(Objid what, Objid this, const char *verbname,
	     Parse_Info *pi, int match_options)
{
    Objid	oid;

    if (!valid(what) || !valid(this))
	return 0;

    pi->this = this;
    for (oid = what; oid != NOTHING; oid = obj_parent(oid)) {
	Program *program = get_program_and_pi(oid, verbname,
					      pi, match_options);

	if (program)
	    return program;
    }
    return 0;
}

/* Command parsing */

Var
parse_into_wordlist(const char *command)
{
    int		argc, i;
    char      **argv;
    Var		args;
    char       *s = str_dup(command);

    argv = parse_into_words(s, &argc);
    args = new_list(argc);
    for (i = 1; i <= argc; i++) {
	args.v.list[i].type = TYPE_STR;
	args.v.list[i].v.str = str_dup(argv[i - 1]);
    }
    free_str(s);
    return args;
}

int
parse_command(const char *command, Parse_Info *pi)
{
    const char *argstr;
    char       *buf;
    const char *verb;
    int		argc;
    char      **argv;
    int		pstart, pend, dlen;
    int		i;

    while (*command == ' ')
	command++;
    switch (*command) {
      case '"':
	verb = "say";
	goto finish_specials;
      case ':':
	verb = "emote";
	goto finish_specials;
      case ';':
	verb = "eval";
	goto finish_specials;
	
      finish_specials:
	argstr = command + 1;
	buf = (char *) mymalloc(strlen(argstr) + strlen(verb) + 2,
				M_STRING);
	strcpy(buf, verb);
	strcat(buf, " ");
	strcat(buf, argstr);
	break;
	
      default:
	buf = str_dup(command);
	{			/* Skip past even complexly-quoted verbs */
	    int	in_quotes = 0;
	    
	    argstr = command;
	    while (*argstr && (in_quotes || *argstr != ' ')) {
		char c = *(argstr++);

		if (c == '"')
		    in_quotes = !in_quotes;
		else if (c == '\\'  &&  *argstr)
		    argstr++;
	    }
	}
	while (*argstr == ' ')
	    argstr++;
	break;
    }
    argv = parse_into_words(buf, &argc);

    if (argc == 0) {
	free_str(buf);
	return 0;
    }

    pi->argstr = str_dup(argstr);
    pi->verb = str_dup(argv[0]);

    pi->args = new_list(argc - 1);
    for (i = 1; i < argc; i++) {
	pi->args.v.list[i].type = TYPE_STR;
	pi->args.v.list[i].v.str = str_dup(argv[i]);
    }

    /*
     * look for a preposition
     */
    if (argc > 1) {
	pi->prep = prep_find(argc - 1, argv + 1, &pstart, &pend);
	if (pi->prep == FAILED_MATCH) {
	    pi->prep = NOTHING;
	    pstart = argc;
	    pend = argc;
	} else {
	    pstart++;
	    pend++;
	}
    } else {
	pi->prep = NOTHING;
	pstart = argc;
	pend = argc;
    }

    /*
     * if there's a preposition,
     * find the iobj & dobj around it, if any
     */
    if (pi->prep != NOTHING) {
	pi->prepstr = build_string(pend - pstart + 1, argv + pstart);
	pi->iobjstr = build_string(argc - (pend + 1), argv + (pend + 1));
	pi->iobj = match_object(pi->player, pi->iobjstr);
    } else {
	pi->prepstr = str_dup("");
	pi->iobjstr = str_dup("");
	pi->iobj = NOTHING;
    }

    dlen = pstart - 1;
    if (dlen == 0) {
	pi->dobjstr = str_dup("");
	pi->dobj = NOTHING;
    } else {
	pi->dobjstr = build_string(dlen, argv + 1);
	pi->dobj = match_object(pi->player, pi->dobjstr);
    }

    free_str(buf);

    return 1;
}

char rcsid_parse_cmd[] = "$Id: parse_cmd.c,v 1.10 1992/10/23 23:03:47 pavel Exp $";

/* $Log: parse_cmd.c,v $
 * Revision 1.10  1992/10/23  23:03:47  pavel
 * Added copyright notice.
 *
 * Revision 1.9  1992/10/21  03:02:35  pavel
 * Converted to use new automatic configuration system.
 *
 * Revision 1.8  1992/10/17  20:49:22  pavel
 * Global rename of strdup->str_dup, strref->str_ref, vardup->var_dup, and
 * varref->var_ref.
 *
 * Revision 1.7  1992/09/14  17:30:26  pjames
 * Moved db_modification code to db modules.
 *
 * Revision 1.6  1992/08/31  22:32:16  pjames
 * Changed some `char *'s to `const char *' and fixed code accordingly.
 *
 * Revision 1.5  1992/08/28  16:15:47  pjames
 * Changed myfree(*, M_STRING) to free_str(*).
 * Changed some strref's to strdup.
 *
 * Revision 1.4  1992/08/20  23:49:12  pavel
 * Moved #define MAXARGS from config.h to here, renamed to MAXWORDS.  Also
 * changed the file name from `parse_command.c' to `parse_cmd.c', to make it
 * easier to use on 14-character-file-name systems.
 *
 * Revision 1.3  1992/08/10  17:35:34  pjames
 * Added find_verb() which doesn't fill in a Parse_Info, just returns
 * information about the verb.
 *
 * Revision 1.2  1992/07/21  00:04:53  pavel
 * Added rcsid_<filename-root> declaration to hold the RCS ident. string.
 *
 * Revision 1.1  1992/07/20  23:23:12  pavel
 * Initial RCS-controlled version.
 */
