/******************************************************************************
  Copyright (c) 1992 Xerox Corporation.  All rights reserved.
  Portions of this code were written by Stephen White, aka ghond.
  Use and copying of this software and preparation of derivative works based
  upon this software are permitted.  Any distribution of this software or
  derivative works must comply with all applicable United States export
  control laws.  This software is made available AS IS, and Xerox Corporation
  makes no warranty about the software, its performance or its conformity to
  any specification.  Any person obtaining a copy of this software is requested
  to send their name and post office or electronic mail address to:
    Pavel Curtis
    Xerox PARC
    3333 Coyote Hill Rd.
    Palo Alto, CA 94304
    Pavel@Xerox.Com
 *****************************************************************************/

#include "my-stdarg.h"
#include "my-stdio.h"
#include "my-stdlib.h"
#include "my-string.h"
#include "my-time.h"

#include "ast.h"
#include "config.h"
#include "db.h"
#include "db_property.h"
#include "exceptions.h"
#include "list.h"
#include "log.h"
#include "match.h"
#include "options.h"
#include "parse_cmd.h"
#include "quota.h"
#include "ref_count.h"
#include "server.h"
#include "storage.h"
#include "streams.h"
#include "structures.h"
#include "tasks.h"
#include "unparse.h"
#include "parser.h" /* parse_program */
#include "utils.h"

Object	**objects;
int	nobjects = 0;
static int	max_objects = 0;
Var		players_list;

static void	write_object(Objid);
static void	write_programs(Objid, Verbdef *, char *);
static int	read_object(FILE *f);
static Objid	read_objid(FILE *f);
static void	write_objid(Objid oid);


static unsigned	alloc_num[Sizeof_Memory_Type];
#ifdef USE_GNU_MALLOC
static unsigned alloc_size[Sizeof_Memory_Type],
                alloc_real_size[Sizeof_Memory_Type];
#endif

void *	 
mymalloc(unsigned size, Memory_Type type)
{
    void       *memptr;
    char	msg[100];

    if (size == 0)		/* For queasy systems */
	size = 1;

    memptr = (void *) malloc(size
#ifdef FIND_LEAKS
		             + 8
#endif
		             );
    if (!memptr) {
	sprintf(msg, "memory allocation (size %u) failed!", size);
	panic(msg);
    }

#ifdef FIND_LEAKS
    *((int *) memptr) = type;
#endif

    alloc_num[type]++;
#ifdef USE_GNU_MALLOC
    {
	extern unsigned		malloc_real_size(void *ptr);
	extern unsigned		malloc_size(void *ptr);

	alloc_size[type] += malloc_size(memptr);
	alloc_real_size[type] += malloc_real_size(memptr);
    }
#endif

#ifdef FIND_LEAKS
    return ((char *) memptr) + 8;
#else
    return memptr;
#endif
}

const char *
str_ref(const char *s)
{
    addref(s);
    return s;
}

char *
str_dup(const char *s)
{
    char	*r;

    if (s == 0 || *s == '\0') {
	r = (char *) mymalloc(1, M_STRING);
	*r = '\0';
    } else {
	r = (char *) mymalloc(strlen(s) + 1, M_STRING);
	strcpy(r, s);
    }
    return  r;
}

void
myfree(void *ptr, Memory_Type type)
{
#ifdef FIND_LEAKS
    ptr = ((char *) ptr) - 8;

    if (*((int *) ptr) != type)
        abort();
#endif

    alloc_num[type]--;
#ifdef USE_GNU_MALLOC
    {
	extern unsigned		malloc_real_size(void *ptr);
	extern unsigned		malloc_size(void *ptr);

	alloc_size[type] -= malloc_size(ptr);
	alloc_real_size[type] -= malloc_real_size(ptr);
    }
#endif

    free(ptr);
}

#ifdef USE_GNU_MALLOC
    struct mstats_value {
	int blocksize;
	int nfree;
	int nused;
    };

    extern struct mstats_value malloc_stats(int size);
#endif

void
free_str(const char *s)
{
    if (delref(s) == 0)
	myfree((void *)s, M_STRING);
}

Var
memory_usage(void)
{
    Var		r;

#ifdef USE_GNU_MALLOC
    int		nsizes, i;

    /* Discover how many block sizes there are. */
    for (nsizes = 0; ; nsizes++) {
	struct mstats_value	v;

	v = malloc_stats(nsizes);
	if (v.blocksize <= 0)
	    break;
    }

    /* Get all of the allocation out of the way before getting the stats. */
    r = new_list(nsizes);
    for (i = 1; i <= nsizes; i++)
	r.v.list[i] = new_list(3);

    for (i = 0; i < nsizes; i++) {
	struct mstats_value	v;
	Var			l;

	v = malloc_stats(i);
	l = r.v.list[i + 1];
	l.v.list[1].type = l.v.list[2].type = l.v.list[3].type = TYPE_NUM;
	l.v.list[1].v.num = v.blocksize;
	l.v.list[2].v.num = v.nused;
	l.v.list[3].v.num = v.nfree;
    }
#else
    r = new_list(0);
#endif

    return r;
}

Object *
retrieve(Objid id)
{
    if (id < 0 || id >= nobjects)
	return 0;
    else
	return objects[id];
}

static void
ensure_room_for_new_object(void)
{
    if (max_objects == 0) {
	max_objects = 100;
	objects = (Object **) mymalloc(max_objects * sizeof(Object *),
				       M_OBJECT_TABLE);
    }
    if (nobjects >= max_objects) {
	int	i;
	Object	**new;

	new = (Object **) mymalloc(max_objects * 2 * sizeof(Object *),
				   M_OBJECT_TABLE);
	for (i = 0; i < max_objects; i++)
	    new[i] = objects[i];
	myfree((void *) objects, M_OBJECT_TABLE);
	objects = new;
	max_objects *= 2;
    }
}

Object *
new_object(void)
{
    Object	*r;

    ensure_room_for_new_object();
    r = objects[nobjects] = (Object *) mymalloc(sizeof(Object), M_OBJECT);
    r->id = nobjects;
    nobjects++;
    return r;
}

void
free_object(Objid oid)
{
    myfree((void *) objects[oid], M_OBJECT);
    objects[oid] = 0;
}

int
db_size()
{
    return (nobjects);
}

void
reset_db_size()
{
    Objid	oid;

    for (oid = nobjects - 1; oid >= 0  &&  !valid(oid); oid--)
	;
    nobjects = oid + 1;
}

static int
validate_hierarchies()
{
    Objid	oid, log_oid;
    Objid	size = db_size();
    int		broken = 0;
    int		fixed_nexts = 0;

    logf("VALIDATING the object hierarchies ...\n");

#   define PROGRESS_INTERVAL 10000
#   define MAYBE_LOG_PROGRESS					\
    {								\
        if (oid == log_oid) {					\
	    log_oid += PROGRESS_INTERVAL;			\
	    logf("VALIDATE: Done through #%d ...\n", oid);	\
	}							\
    }

    logf("VALIDATE: Phase 1: Check for invalid objects ...\n");
    for (oid = 0, log_oid = PROGRESS_INTERVAL; oid < size; oid++) {
	Object *o = retrieve(oid);

	MAYBE_LOG_PROGRESS;
	if (o) {
	    if (o->location == NOTHING  &&  o->next != NOTHING) {
		o->next = NOTHING;
		fixed_nexts++;
	    }

#	    define CHECK(field, name) 					\
	    {								\
	        if (o->field != NOTHING && !retrieve(o->field)) {	\
		    errlogf("VALIDATE: #%d.%s = #%d <invalid> ... fixed.\n", \
			   oid, name, o->field);			\
		    o->field = NOTHING;				  	\
		}							\
	    }

	    CHECK(parent, "parent");
	    CHECK(child, "child");
	    CHECK(sibling, "sibling");
	    CHECK(location, "location");
	    CHECK(contents, "contents");
	    CHECK(next, "next");

#	    undef CHECK
	}
    }

    if (fixed_nexts != 0)
	errlogf("VALIDATE: Fixed %d should-be-null next pointer(s) ...\n",
	       fixed_nexts);

    logf("VALIDATE: Phase 2: Check for cycles ...\n");
    for (oid = 0, log_oid = PROGRESS_INTERVAL; oid < size; oid++) {
	Object	*o = retrieve(oid);

	MAYBE_LOG_PROGRESS;
	if (o) {
#	    define CHECK(start, field, name)				\
	    {								\
	        Objid	oid2 = start;					\
		int	count = 0;					\
	        for (; oid2 != NOTHING; oid2 = retrieve(oid2)->field) {	\
		    if (++count > size)	{				\
			errlogf("VALIDATE: Cycle in `%s' chain of #%d\n",\
			       name, oid);				\
			broken = 1;					\
			break;						\
		    }							\
		}							\
	    }

	    CHECK(o->parent, parent, "parent");
	    CHECK(o->child, sibling, "child");
	    CHECK(o->location, location, "location");
	    CHECK(o->contents, next, "contents");

#	    undef CHECK
	}
    }

    if (broken)			/* Can't continue if cycles found */
	return 0;

    logf("VALIDATE: Phase 3: Check for inconsistencies ...\n");
    for (oid = 0, log_oid = PROGRESS_INTERVAL; oid < size; oid++) {
	Object	*o = retrieve(oid);

	MAYBE_LOG_PROGRESS;
	if (o) {
#	    define CHECK(up, up_name, down, down_name, across)		\
	    {								\
		Objid	up = o->up;					\
		Objid	oid2;						\
									\
		/* Is oid in its up's down list? */			\
		if (up != NOTHING) {					\
		    for (oid2 = retrieve(up)->down;			\
			 oid2 != NOTHING;				\
			 oid2 = retrieve(oid2)->across) {		\
			if (oid2 == oid) /* found it */			\
			    break;					\
		    }							\
		    if (oid2 == NOTHING) { /* didn't find it */		\
			errlogf("VALIDATE: #%d not in %s (#%d)'s %s list.\n", \
			       oid, up_name, up, down_name);	        \
			broken = 1;					\
		    }							\
		}							\
	    }

	    CHECK(parent, "parent", child, "child", sibling);
	    CHECK(location, "location", contents, "contents", next);

#	    undef CHECK

#	    define CHECK(up, down, down_name, across)			\
	    {								\
		Objid	oid2;						\
									\
		for (oid2 = o->down;					\
		     oid2 != NOTHING;					\
		     oid2 = retrieve(oid2)->across) {			\
		    if (retrieve(oid2)->up != oid) {			\
			errlogf(						\
			    "VALIDATE: #%d erroneously on #%d's %s list.\n", \
			    oid2, oid, down_name);			\
			broken = 1;					\
		    }							\
		}							\
	    }

	    CHECK(parent, child, "child", sibling);
	    CHECK(location, contents, "contents", next);
	    
#	    undef CHECK
	}
    }
    
    logf("VALIDATING the object hierarchies ... finished.\n");
    return !broken;
}

static Exception	dump_failed;
static FILE	       *dump_file;

void
db_printf(const char *format, ...)
{
    va_list	args;

    va_start(args, format);
    if (vfprintf(dump_file, format, args) < 0)
	RAISE(dump_failed, 0);
    va_end(args);
}

int
db_write(FILE *f)
{
    Objid	oid;
    Verbdef	*v;
    int		i;
    volatile int nprogs = 0;
    volatile int success = 1;

    dump_file = f;

    for (oid = 0; oid < nobjects; oid++) {
	if (!valid(oid))
	    continue;

	for (v = retrieve(oid)->verbdefs; v; v = v->next)
	    if (v->program)
		nprogs++;
    }
    
    TRY
	db_printf("%d\n%d\n%d\n%d\n", nobjects, nprogs, 0,
		  players_list.v.list[0].v.num);
	for (i = 0; i < players_list.v.list[0].v.num; i++)
	    write_objid(players_list.v.list[i + 1].v.obj);
	for (oid = 0; oid < nobjects; oid++)
	    write_object(oid);
	for (oid = 0; oid < nobjects; oid++) {
	    char	buf[20];
	    
	    if (!valid(oid))
		continue;
	    sprintf(buf, "#%d", oid);
	    write_programs(obj_parent(oid),
			   retrieve(oid)->verbdefs, buf);
	}
	write_task_queue(f);
        write_active_connections(f);
    EXCEPT (dump_failed)
	success = 0;
    ENDTRY;

    return success;
}

extern Verbdef *find_nth_verbdef(Objid, unsigned);

int
db_read(FILE *f)
{
    Objid	no, oid;
    int		nprogs;		/* number of programs */
    int		nplayers;
    int		vnum;
    Verbdef	*v;
    int		dummy;
    int		i;
    char	c;

    if (fscanf(f, "%d\n%d\n%d\n%d\n", &no, &nprogs, &dummy, &nplayers) != 4) {
	errlogf("DB_READ: Bad header\n");
	return -1;
    }

    players_list = new_list(nplayers);
    for (i = 1; i <= nplayers; i++) {
	players_list.v.list[i].type = TYPE_OBJ;
	players_list.v.list[i].v.obj = read_objid(f);
    }

    logf("LOADING: Reading %d objects...\n", no);
    for (i = 1; i <= no; i++) {
	if (read_object(f)) {
	    errlogf("DB_READ: Bad object, #%d.\n", i - 1);
	    return -1;
	}
	if (i % 10000 == 0  ||  i == no)
	    logf("LOADING: Done reading %d objects...\n", i);
    }

    if (!validate_hierarchies()) {
	errlogf("DB_READ: Errors found in object hierarchies; read aborted.\n");
	return -1;
    }

    logf("LOADING: Reading %d MOO programs...\n", nprogs);
    for (i = 1; i <= nprogs; i++) {
	if (fscanf(f, "#%d:%d", &oid, &vnum) != 2) {
	    errlogf("DB_READ: Bad program header, i = %d.\n", i);
	    return -1;
	}
	c = fgetc(f);
	if (c != '\n') {
	    errlogf("DB_READ: Bad program header char: '%c', i = %d.\n", c, i);
	    return -1;
	}
	if (!valid(oid)) {
	    errlogf("DB_READ: Verb programmed on an non-existant object: oid = %d, i = %d.\n",
		   oid, i);
	    return -1;
	}
	if (!(v = find_nth_verbdef(oid, vnum + 1))) {
	    errlogf("DB_READ: Unknown verb index: #%d:%d, i = %d.\n",
		   oid, vnum, i);
	    return -1;
	}
	if (read_program(f, &v->program) < 0) {
	    errlogf("DB_READ: Unparsable program #%d:%d, i = %d.\n",
		   oid, vnum, i);
	    return -1;
	}
	if (i % 5000 == 0  ||  i == nprogs)
	    logf("LOADING: Finished reading %d programs...\n", i);
    }

    logf("LOADING: Reading forked and suspended tasks...\n");
    if (read_task_queue(f) < 0) {
	errlogf("DB_READ: Can't read task queue.\n");
	return -1;
    }
    logf("LOADING: Reading list of formerly active connections...\n");
    if (read_active_connections(f) < 0) {
	errlogf("DB_READ: Can't read active connections.\n");
	return -1;
    }
    return 0;
}

#define RS_BUFLEN	1024

const char *
read_string(FILE *f)
{
    static Stream      *str = 0;
    static char		buffer[RS_BUFLEN];
    int			len, used_stream = 0;

    if (str == 0)
	str = new_stream(1024);

  try_again:
    fgets(buffer, RS_BUFLEN, f);
    len = strlen(buffer);
    if (len == RS_BUFLEN - 1  &&  buffer[len - 1] != '\n') {
	stream_add_string(str, buffer);
	used_stream = 1;
	goto try_again;
    }

    if (buffer[len - 1] == '\n')
	buffer[len - 1] = '\0';

    if (used_stream) {
	stream_add_string(str, buffer);
	return reset_stream(str);
    } else
	return buffer;
}

void
write_string(const char *s)
{
    if (s)
	db_printf("%s\n", s);
    else
	db_printf("\n");
}

int
read_num(FILE *f)
{
    char	s[20];
    char       *p;
    int		i;

    fgets(s, 20, f);
    i = strtol(s, &p, 10);
    if (*p != '\n')
	errlogf("READ_NUM: Bad number: \"%s\"\n", s);
    return i;
}

void
write_num(int n)
{
    db_printf("%d\n", n);
}

static Objid
read_objid(FILE *f)
{
    return read_num(f);
}

static void
write_objid(Objid oid)
{
    write_num(oid);
}

static void
read_verbdef(FILE *f, Verbdef *v)
{
    v->name = str_dup(read_string(f));
    v->owner = read_objid(f);
    v->perms = read_num(f);
    v->prep = read_num(f);
    v->next = 0;
    v->program = 0;
}

static void
write_verbdef(Verbdef *v)
{
    write_string(v->name);
    write_objid(v->owner);
    write_num(v->perms);
    write_num(v->prep);
}

static Propdef
read_propdef(FILE *f)
{
    char *name = str_dup(read_string(f));
    return new_propdef(name);
}

static void
write_propdef(Propdef *p)
{
    write_string(p->name);
}

Var
read_var(FILE *f)
{
    Var		r;
    int		i, l = read_num(f);

    r.type = (var_type) l;
    switch (l) {
      case TYPE_CLEAR:
      case (int) TYPE_ANY:
	break;
      case TYPE_STR:
	r.v.str = str_dup(read_string(f));
	break;
      case TYPE_OBJ:
      case TYPE_ERR:
      case TYPE_NUM:
	r.v.num = read_num(f);
	break;
      case TYPE_LIST:
	l = read_num(f);
	r = new_list(l);
	for (i = 0; i < l; i++)
	    r.v.list[i + 1] = read_var(f);
	break;
      default:
	errlogf("READ_VAR: Unknown type (%d) at DB file pos. %ld\n",
	       l, ftell(f));
	r.type = TYPE_NUM;
	r.v.num = 0;
	break;
    }
    return r;
}

static void
read_propval(FILE *f, Pval  *p)
{
    p->var = read_var(f);
    p->owner = read_objid(f);
    p->perms = read_num(f);
}

void
write_var(Var v)
{
    int		i;

    write_num((int) v.type);
    switch ((int) v.type) {
      case TYPE_CLEAR:
      case (int) TYPE_ANY:
	break;
      case TYPE_STR:
	write_string(v.v.str);
	break;
      case TYPE_OBJ:
      case TYPE_ERR:
      case TYPE_NUM:
	write_num(v.v.num);
	break;
      case TYPE_LIST:
	write_num(v.v.list[0].v.num);
	for (i = 0; i < v.v.list[0].v.num; i++)
	    write_var(v.v.list[i + 1]);
	break;
    }
}

static void
write_propval(Pval  *p)
{
    write_var(p->var);
    write_objid(p->owner);
    write_num(p->perms);
}

static int
read_object(FILE *f)
{
    Objid	oid;
    Object     *o;
    Verbdef    *v, **prevv;
/*    Propdef   **prevp;*/
    int		i;
    Propid	pid = 0;
    int		nprops;
    char	s[20];

    if (fscanf(f, "#%d", &oid) != 1  ||  oid != nobjects)
	return -1;
    fgets(s, 20, f);

    if (strcmp(s, " recycled\n") == 0) {
	ensure_room_for_new_object();
	objects[nobjects++] = 0;
	return 0;
    } else if (strcmp(s, "\n") != 0)
	return -1;

    o = new_object();

    o->name = str_dup(read_string(f));
    (void) read_string(f);	/* discard old handles string */
    o->flags = read_num(f);

    o->owner = read_objid(f);

    o->location = read_objid(f);
    o->contents = read_objid(f);
    o->next = read_objid(f);

    o->parent = read_objid(f);
    o->child = read_objid(f);
    o->sibling = read_objid(f);

    o->verbdefs = 0;
    o->propdefs.cur_length = 0;
    o->propdefs.max_length = 0;
    o->propdefs.l = 0;

    prevv = &(o->verbdefs);
    for (i = read_num(f); i > 0; i--) {
	v = (Verbdef *) mymalloc(sizeof(Verbdef), M_VERBDEF);
	read_verbdef(f, v);
	*prevv = v;
	prevv = &(v->next);
    }

    if ((i = read_num(f)) != 0) {
	o->propdefs.l = mymalloc(i * sizeof(Propdef), M_PROPDEF);
	o->propdefs.cur_length = i;
	o->propdefs.max_length = i;
	for (i = 0; i < o->propdefs.cur_length; i++)
	    o->propdefs.l[i] = read_propdef(f);
    }
	
    nprops = read_num(f);
    if (nprops) {
	o->propval = (Pval *) mymalloc(nprops * sizeof(Pval), M_PVAL);
    } else {
	o->propval = 0;
    }

    for (pid = 0; pid < nprops; pid++) {
	read_propval(f, o->propval + pid);
    }

    return 0;
}

static void
receiver(void *data, const char *line)
{
    db_printf("%s\n", line);
}

void
write_program(Program *program)
{
    list_prog(program, receiver, 0, 1, 0, MAIN_VECTOR);
    db_printf(".\n");
}

void
write_forked_program(Program *program, unsigned f_index)
{
    list_prog(program, receiver, 0, 1, 0, f_index);
    db_printf(".\n");
}

struct parser_state {
    FILE       *f;
    char	prev_char;
};

static void
my_error(void *data, const char *msg)
{
    errlogf("PARSER: %s\n", msg);
}

static int
my_getc(void *data)
{
    struct parser_state	*state = (struct parser_state *) data;
    int		c;

    c = fgetc(state->f);
    if (c == '.' && state->prev_char == '\n') {
	/* end-of-verb marker in DB */
	c = fgetc(state->f);		/* skip next newline */
	return EOF;
    }
    if (c == EOF)
	my_error(data, "Unexpected EOF");
    state->prev_char = c;
    return c;
}

static void
my_ungetc(void *data, char c)
{
    struct parser_state	*state = (struct parser_state *) data;

    ungetc(c, state->f);
}

static Parser_Client    parser_client = { my_error, my_ungetc, my_getc };

int
read_program(FILE *f, Program **program)
{
    struct parser_state	state;

    state.f = f;
    state.prev_char = '\n';
    *program = parse_program(parser_client, (void *) &state);
    if (!*program)
	return -1;
    return 0;
}

static void
write_programs(Objid what, Verbdef *vstart, char *whatname)
{
    Verbdef    *v;
    int		vcount = 0;

    for (v = vstart; v; v = v->next) {
	if (v->program) {
	    db_printf("%s:%d\n", whatname, vcount);
	    write_program(v->program);
	}
	vcount++;
    }
}

static void
write_object(Objid oid)
{
    Object     *o;
    Verbdef    *v;
    Propid	pid = 0;
    unsigned    loop;
    int		nverbdefs = 0,/* npropdefs = 0, */nprops = 0;

    o = retrieve(oid);
    if (!valid(oid)) {
	db_printf("#%d recycled\n", oid);
	return;
    }
    db_printf("#%d\n", oid);
    write_string(obj_name(oid));
    write_string("");	/* placeholder for old handles string */
    write_num(obj_flags(oid));

    write_objid(obj_owner(oid));

    write_objid(obj_location(oid));
    write_objid(obj_contents(oid));
    write_objid(obj_next_content(oid));

    write_objid(obj_parent(oid));
    write_objid(obj_child(oid));
    write_objid(obj_sibling(oid));

    for (v = o->verbdefs; v; v = v->next) {
	nverbdefs++;
    }
    write_num(nverbdefs);
    for (v = o->verbdefs; v; v = v->next) {
	write_verbdef(v);
    }

    write_num(o->propdefs.cur_length);
    for (loop = 0; loop < o->propdefs.cur_length; loop++)
	write_propdef(&o->propdefs.l[loop]);
    
    nprops = count_properties(oid);
    write_num(nprops);
    for (pid = 0; pid < nprops; pid++) {
	write_propval(o->propval + pid);
    }
    return;
}

int
is_wizard(Objid oid)
{
    return valid(oid)  &&  (obj_flags(oid) & (1 << FLAG_WIZARD)) != 0;
}

int
is_programmer(Objid oid)
{
    return valid(oid)  &&  (obj_flags(oid) & (1 << FLAG_PROGRAMMER)) != 0;
}

int
is_player(Objid oid)
{
    return valid(oid)  &&  (obj_flags(oid) & (1 << FLAG_PLAYER)) != 0;
}

/* from lambda/sym_table.c -- then heavily modified */

void
write_rt_env(const char **var_names, Var *rt_env, unsigned size)
{
    unsigned	i;

    db_printf("%d variables\n", size);
    for (i = 0; i < size; i++) {
	write_string(var_names[i]);
	write_var(rt_env[i]);
    }
}

int
read_rt_env(FILE *f, const char ***old_names, Var **rt_env, int *old_size)
{
    unsigned	i;

    if (fscanf(f, "%d variables\n", old_size) != 1) {
	errlogf("READ_ENV: Bad count.\n");
	return -1;
    }

    *old_names = (const char **) mymalloc((*old_size) * sizeof(char *),
					  M_NAMES);
    *rt_env = new_rt_env(*old_size);

    for (i = 0; i < *old_size; i++) {
	(*old_names)[i] = str_dup(read_string(f));
	(*rt_env)[i] = read_var(f);
    }
    return 0;
}

Var *
reorder_rt_env(Var *old_rt_env, const char **old_names,
	       int old_size, Program *prog)
{
    /* reorder old_rt_env, which is aligned according to old_names, 
       to align to prog->var_names -- return the new rt_env
       after freeing old_rt_env and old_names */

    unsigned size = prog->num_var_names;
    Var *rt_env = new_rt_env(size);

    unsigned i;

    for (i = 0; i < size; i++) {
	int	slot;

	for (slot = 0; slot < old_size; slot++) {
	    if (mystrcasecmp(old_names[slot], prog->var_names[i]) == 0)
		break;
	}
				
	if (slot < old_size)
	    rt_env[i] = var_ref(old_rt_env[slot]);
    }

    free_rt_env(old_rt_env, old_size);
    for (i = 0; i < old_size; i++)
	free_str(old_names[i]);
    myfree((void *) old_names, M_NAMES);

    return rt_env;
}


char rcsid_storage[] = "$Id: storage.c,v 1.16 1992/10/23 23:03:47 pavel Exp $";

/* $Log: storage.c,v $
 * Revision 1.16  1992/10/23  23:03:47  pavel
 * Added copyright notice.
 *
 * Revision 1.15  1992/10/21  03:02:35  pavel
 * Converted to use new automatic configuration system.
 *
 * Revision 1.14  1992/10/17  20:52:37  pavel
 * Global rename of strdup->str_dup, strref->str_ref, vardup->var_dup, and
 * varref->var_ref.
 *
 * Revision 1.13  1992/09/14  18:38:42  pjames
 * Updated #includes.  Moved rcsid to bottom.
 *
 * Revision 1.12  1992/09/14  17:41:16  pjames
 * Moved db_modification code to db modules.
 *
 * Revision 1.11  1992/09/03  16:26:29  pjames
 * Added TYPE_CLEAR handling.
 * Changed property definition manipulating to work with arrays.
 *
 * Revision 1.10  1992/08/31  22:25:27  pjames
 * Changed some `char *'s to `const char *'
 *
 * Revision 1.9  1992/08/28  16:21:13  pjames
 * Changed vardup to varref.
 * Changed myfree(*, M_STRING) to free_str(*).
 * Added `strref()' and `free_str()'
 *
 * Revision 1.8  1992/08/21  00:42:18  pavel
 * Renamed include file "parse_command.h" to "parse_cmd.h".
 *
 * Changed to conditionalize on the option USE_GNU_MALLOC instead of
 * USE_SYSTEM_MALLOC.
 *
 * Removed use of worthless constant DB_INITIAL_SIZE, defined in config.h.
 *
 * Revision 1.7  1992/08/14  00:00:45  pavel
 * Converted to a typedef of `var_type' = `enum var_type'.
 *
 * Revision 1.6  1992/08/10  16:52:45  pjames
 * Updated #includes.
 *
 * Revision 1.5  1992/07/30  21:23:10  pjames
 * Casted malloc to (void *) because of a problem with some system.
 *
 * Revision 1.4  1992/07/27  19:05:18  pjames
 * Removed a debugging statement.
 *
 * Revision 1.3  1992/07/27  18:17:41  pjames
 * Changed name of ct_env to var_names and M_CT_ENV to M_NAMES.
 *
 * Revision 1.2  1992/07/21  00:06:38  pavel
 * Added rcsid_<filename-root> declaration to hold the RCS ident. string.
 *
 * Revision 1.1  1992/07/20  23:23:12  pavel
 * Initial RCS-controlled version.
 */
