/******************************************************************************
  Copyright (c) 1992 Xerox Corporation.  All rights reserved.
  Portions of this code were written by Stephen White, aka ghond.
  Use and copying of this software and preparation of derivative works based
  upon this software are permitted.  Any distribution of this software or
  derivative works must comply with all applicable United States export
  control laws.  This software is made available AS IS, and Xerox Corporation
  makes no warranty about the software, its performance or its conformity to
  any specification.  Any person obtaining a copy of this software is requested
  to send their name and post office or electronic mail address to:
    Pavel Curtis
    Xerox PARC
    3333 Coyote Hill Rd.
    Palo Alto, CA 94304
    Pavel@Xerox.Com
 *****************************************************************************/

#include "my-stdio.h"

#include "ast.h"
#include "config.h"
#include "exceptions.h"
#include "log.h"
#include "storage.h"
#include "structures.h"
#include "sym_table.h"
#include "utils.h"

Names *
new_names(unsigned max_size)
{
    Names      *names = (Names *) mymalloc(sizeof(Names), M_NAMES);
    names->names = (const char **) mymalloc(sizeof(char *) * max_size,
					    M_NAMES);

    names->max_size = max_size;
    names->size = 0;

    return names;
}

Names *
new_builtin_names(void)
{
    static Names       *builtins = 0;

    if (builtins == 0) {
	builtins = new_names(FIRST_USER_SLOT);
	builtins->size = FIRST_USER_SLOT;

	builtins->names[SLOT_NUM] = str_dup("NUM");
	builtins->names[SLOT_OBJ] = str_dup("OBJ");
	builtins->names[SLOT_STR] = str_dup("STR");
	builtins->names[SLOT_LIST] = str_dup("LIST");
	builtins->names[SLOT_ERR] = str_dup("ERR");
	builtins->names[SLOT_PLAYER] = str_dup("player");
	builtins->names[SLOT_THIS] = str_dup("this");
	builtins->names[SLOT_CALLER] = str_dup("caller");
	builtins->names[SLOT_VERB] = str_dup("verb");
	builtins->names[SLOT_ARGS] = str_dup("args");
	builtins->names[SLOT_ARGSTR] = str_dup("argstr");
	builtins->names[SLOT_DOBJ] = str_dup("dobj");
	builtins->names[SLOT_DOBJSTR] = str_dup("dobjstr");
/*  taken out to match code_elt.c
	builtins->names[SLOT_PREP] = str_dup("prep");
*/
	builtins->names[SLOT_PREPSTR] = str_dup("prepstr");
	builtins->names[SLOT_IOBJ] = str_dup("iobj");
	builtins->names[SLOT_IOBJSTR] = str_dup("iobjstr");
    }
    return copy_names(builtins);
}

Names *
copy_names(Names *old)
{
    Names      *new = new_names(old->size);
    unsigned	i;

    new->size = old->size;
    for (i = 0; i < new->size; i++)
	new->names[i] = str_ref(old->names[i]);

    return new;
}

int
find_name(Names *names, const char *str)
{
    unsigned	i;

    for (i = 0; i < names->size; i++)
	if (!mystrcasecmp(names->names[i], str))
	    return i;
    return -1;
}

unsigned
find_or_add_name(Names **names, const char *str)
{
  unsigned i;

  for (i = 0; i < (*names)->size; i++)
    if (!mystrcasecmp((*names)->names[i], str)) {/* old name */
      return i;
    }
  
  if ((*names)->size == (*names)->max_size) {
    unsigned	old_max = (*names)->max_size;
    Names      *new = new_names(old_max * 2);
    unsigned	i;

    for (i = 0; i < old_max; i++)
      new->names[i] = (*names)->names[i];
    new->size = old_max;
    myfree((void *) (*names)->names, M_NAMES);
    myfree((void *) *names, M_NAMES);
    *names = new;
  }
  (*names)->names[(*names)->size] = str_ref(str);
  return (*names)->size++;
}

void
free_names(Names *names)
{
    unsigned	i;

    for (i = 0; i < names->size; i++)
	free_str(names->names[i]);
    myfree((void *) names->names, M_NAMES);
    myfree((void *) names, M_NAMES);
}

Env *
new_env(unsigned size)
{
    Env	       *new = (Env *) mymalloc(sizeof(Env) +
				       (size - 1) * sizeof(Var),
				       M_ENV);
    unsigned	i;

    new->size = size;
    for (i = 0; i < size; i++)
	new->vars[i].type = TYPE_ANY;

    return new;
}

Env *
copy_env(Env *old)
{
    Env	       *new = new_env(old->size);
    unsigned	i;

    for (i = 0; i < old->size; i++)
	new->vars[i] = var_ref(old->vars[i]);

    return new;
}

void
free_env(Env *env)
{
    unsigned	i;

    for (i = 0; i < env->size; i++)
	free_var(env->vars[i]);
    myfree((void *) env, M_ENV);
}

Var
env_ref(Env *env, unsigned slot)
{
    if (slot < env->size)
	return var_ref(env->vars[slot]);
    else {
	Var	v;
	
	panic("ENV_REF: Slot index out of bounds.");
	v.type = TYPE_NUM;
	v.v.num = 0;
	return v;
    }
}

void
env_set(Env *env, unsigned slot, Var v)
{
    if (slot < env->size) {
	free_var(env->vars[slot]);
	env->vars[slot] = var_ref(v);
    } else
	panic("ENV_SET: Slot index out of bounds.");
}    

/*
void
write_env(FILE *f, Names *names, Env *env)
{
    unsigned	i;

    if (names->size != env->size)
	panic("WRITE_ENV: Sizes don't match!");

    fprintf(f, "%d variables\n", env->size);
    for (i = 0; i < env->size; i++) {
	write_string(f, names->names[i]);
	write_var(f, env->vars[i]);
    }
}
*/
int
read_env(FILE *f, Names **names, Env **env)
{
    int		count;
    unsigned	i;

    if (fscanf(f, "%d variables\n", &count) != 1) {
	errlogf("READ_ENV: Bad count.\n");
	return -1;
    }

    *names = new_names(count);
    (*names)->size = count;
    *env = new_env(count);

    for (i = 0; i < count; i++) {
	(*names)->names[i] = str_dup(read_string(f));
	(*env)->vars[i] = read_var(f);
    }
    return 0;
}


char rcsid_sym_table[] = "$Id: sym_table.c,v 1.10 1992/10/23 23:03:47 pavel Exp $";

/* $Log: sym_table.c,v $
 * Revision 1.10  1992/10/23  23:03:47  pavel
 * Added copyright notice.
 *
 * Revision 1.9  1992/10/21  03:02:35  pavel
 * Converted to use new automatic configuration system.
 *
 * Revision 1.8  1992/10/17  20:54:04  pavel
 * Global rename of strdup->str_dup, strref->str_ref, vardup->var_dup, and
 * varref->var_ref.
 *
 * Revision 1.7  1992/09/08  21:58:33  pjames
 * Updated #includes.
 *
 * Revision 1.6  1992/08/31  22:24:18  pjames
 * Changed some `char *'s to `const char *'
 *
 * Revision 1.5  1992/08/28  16:07:24  pjames
 * Changed vardup to varref.
 * Changed some strref's to strdup.
 * Changed myfree(*, M_STRING) to free_str(*).
 *
 * Revision 1.4  1992/08/10  16:50:28  pjames
 * Updated #includes.
 *
 * Revision 1.3  1992/07/27  18:23:48  pjames
 * Changed M_CT_ENV to M_NAMES.  Freed some memory that wasn't being
 * freed when growing or freeing a list of names.
 *
 * Revision 1.2  1992/07/21  00:07:06  pavel
 * Added rcsid_<filename-root> declaration to hold the RCS ident. string.
 *
 * Revision 1.1  1992/07/20  23:23:12  pavel
 * Initial RCS-controlled version.
 */
